'''
%--------------------------------------------------------------------------
% This file is part of the TIGRE Toolbox
%
% Copyright (c) 2015, University of Bath and
%                     CERN-European Organization for Nuclear Research
%                     All rights reserved.
%
% License:            Open Source under BSD.
%                     See the full license at
%                     https://github.com/CERN/TIGRE/license.txt
%
% Contact:            tigre.toolbox@gmail.com
% Codes:              https://github.com/CERN/TIGRE/
% Coded by:           MATLAB (original): Manasavee Lohvithee
                      Python: Reuben Lindroos
%--------------------------------------------------------------------------
'''

from __future__ import division
import numpy as np


def Measure_Quality(res_prev, res, QualMeasOpts):
    if QualMeasOpts == 'RMSE':
        N = len(res_prev.ravel())
        diff = res_prev - res
        return np.sqrt(sum(diff.ravel() ** 2) / N)

    if QualMeasOpts == 'CC':
        return np.corrcoef(res_prev.ravel(), res.ravel())

    if QualMeasOpts == 'MSSIM':
        N = len(res_prev.ravel())
        # NOTE: May not be necessary in python

        res_prev = res_prev.ravel()
        res = res.ravel()

        # Compute the mean pixel values of the two images

        mean_res_p = res_prev.mean(axis=0)
        mean_res = res.mean(axis=0)
        if mean_res==0 and mean_res_p==0:
            raise ValueError('Initialising with 0 matrix not valid')
        # Luminance Comparison

        K1 = 0.01  # K1 is a small constant <<1
        d = max(res_prev) - min(res_prev)  # dynamic range of the pixel values
        l = ((2 * mean_res * mean_res_p) + (K1 * d) ** 2) / ((mean_res_p ** 2)
                                                             + (mean_res ** 2) + K1 * d ** 2)

        # Contrast comparison

        K2 = 0.02
        sres_p = res_prev.std(axis=0)
        sres = res.std(axis=0)

        c = ((2 * sres_p * sres) + (K2 * d) ** 2) / ((sres_p ** 2) + (sres ** 2) + K2 * d ** 2)

        # Structure comparison
        diffres_p = res_prev - mean_res_p
        diffres = res - mean_res
        delta = (1 / (N - 1)) * sum(diffres_p * diffres)
        s = (delta + (((K2 * d) ** 2)) / 2) / ((sres_p * sres) + ((K2 * d ** 2) / 2))

        return (1 / N) * l * c * s
    if QualMeasOpts=='UQI':
        res = res.ravel()
        res_prev = res_prev.ravel()
        N=len(res_prev)

        # Mean
        mean_res_p = np.mean(res_prev,dtype=np.float32)
        mean_res = np.mean(res, dtype=np.float32)

        # Variance
        varres_p = np.var(res_prev)
        varres = np.var(res)
        if mean_res==0 and mean_res_p==0:
            raise ValueError('Initialising with 0 matrix not valid')
        # Covariance
        cova = sum((res)-mean_res)*((res_prev) - mean_res_p)/(N-1)
        front = (2*cova)/(varres+varres_p)
        back = (2*mean_res*mean_res_p)/((mean_res**2)+(mean_res_p**2))

        return sum(front * back)
    if QualMeasOpts=='SSD':
        return np.sum((res_prev[:,:,0:res_prev.shape[2]]-res[:,:,0:res.shape[2]])**2)






from __future__ import division
import numpy as np
from tigre.algorithms.iterative_recon_alg import IterativeReconAlg
from tigre.algorithms.iterative_recon_alg import decorator
from tigre.utilities.Ax import Ax
from tigre.utilities.Atb import Atb
import time
import os
import math


class CGLS(IterativeReconAlg):
    __doc__ = (" CGLS_CBCT solves the CBCT problem using the conjugate gradient least\n"
               " squares\n"
               " \n"
               "  CGLS_CBCT(PROJ,GEO,ANGLES,NITER) solves the reconstruction problem\n"
               "  using the projection data PROJ taken over ALPHA angles, corresponding\n"
               "  to the geometry descrived in GEO, using NITER iterations.") + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):
        # Don't precompute V and W.
        kwargs.update(dict(W=None, V=None))
        kwargs.update(dict(blocksize=angles.shape[0]))
        self.log_parameters = False
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)

        if self.log_parameters:
            parameter_history = {}
            iterations = self.niter
            parameter_history['alpha'] = np.zeros([iterations], dtype=np.float32)
            parameter_history['beta'] = np.zeros([iterations], dtype=np.float32)
            parameter_history['gamma'] = np.zeros([iterations], dtype=np.float32)
            parameter_history['q_norm'] = np.zeros([iterations], dtype=np.float32)
            parameter_history['s_norm'] = np.zeros([iterations], dtype=np.float32)
            self.parameter_history = parameter_history

        self.__r__ = self.proj - Ax(self.res, self.geo, self.angles, 'ray-voxel')
        self.__p__ = Atb(self.__r__, self.geo, self.angles)
        p_norm = np.linalg.norm(self.__p__.ravel(), 2)
        self.__gamma__ = p_norm * p_norm

    def reinitialise_cgls(self):
        self.__r__ = self.proj - Ax(self.res, self.geo, self.angles, 'ray-voxel')
        self.__p__ = Atb(self.__r__, self.geo, self.angles)
        p_norm = np.linalg.norm(self.__p__.ravel(), 2)
        self.__gamma__ = p_norm * p_norm

    # Overide
    def run_main_iter(self):
        self.l2l = np.zeros([self.niter], dtype=np.float32)
        for i in range(self.niter):
            if i == 0:
                print("CGLS Algorithm in progress.")
                toc = time.clock()
            if i == 1:
                tic = time.clock()
                print('Esitmated time until completetion (s): ' + str((self.niter - 1) * (tic - toc)))
            q = Ax(self.__p__, self.geo, self.angles, 'ray-voxel')
            q_norm = np.linalg.norm(q.ravel(), 2)
            alpha = self.__gamma__ / (q_norm * q_norm)
            self.res += alpha * self.__p__

            for item in self.__dict__:
                if type(getattr(self, item)) == np.ndarray:
                    if np.isnan(getattr(self, item)).any():
                        raise ValueError('nan found for ' + item + ' at iteraton ' + str(i))

            aux = self.proj - Ax(self.res, self.geo, self.angles, 'ray-voxel')
            self.l2l[i] = np.linalg.norm(aux.ravel(), 2)
            if i > 0 and self.l2l[i] > self.l2l[i - 1]:
                print('re-initialization was called at iter:' + str(i))
                self.res -= alpha * self.__p__
                self.reinitialise_cgls()

            self.__r__ -= alpha * q
            s = Atb(self.__r__, self.geo, self.angles)
            s_norm = np.linalg.norm(s.ravel(), 2)

            gamma1 = s_norm * s_norm
            beta = gamma1 / self.__gamma__
            if self.log_parameters:
                self.parameter_history['alpha'][i] = alpha
                self.parameter_history['beta'][i] = beta
                self.parameter_history['gamma'][i] = self.__gamma__
                self.parameter_history['q_norm'][i] = q_norm
                self.parameter_history['s_norm'][i] = s_norm

            self.__gamma__ = gamma1
            self.__p__ = s + beta * self.__p__


cgls = decorator(CGLS, name='cgls')



from __future__ import division
import numpy as np
#TODO: Implement angularDistance (for blocks)

def order_subsets(angles, blocksize, mode):
    # Parse no blocks
    if blocksize is None or blocksize==1:
        index_alpha=np.arange(angles.shape[0])

        if mode is None or mode =='ordered':
            return angles, index_alpha

        if mode =='random':
            np.random.shuffle(index_alpha)
            return angles[index_alpha], index_alpha
        if mode == 'angularDistance':
            # Sort values according to size
            index_alpha=np.argsort(angles)
            # Save the first values for index_alpha and angles in the new lists
            # (this needs to be done or argmin will return error otherwise)
            new_index = [index_alpha[0]]
            new_angles = [angles[index_alpha[0]]]
            # remove the first values from the original arrays
            angles = np.delete(angles[index_alpha],0)
            index_alpha=np.delete(index_alpha,0)
            # Run through list
            for i in range(len(angles)):
                # Run the comparison
                compangle = new_angles[i]
                angles_min = np.argmin(abs(abs(angles-compangle)-(np.pi/2)))
                # Save the results to the new lists
                new_angles.append(angles[angles_min])
                new_index.append(index_alpha[angles_min])
                # delete old values from the original arrays
                angles=np.delete(angles,angles_min)
                index_alpha=np.delete(index_alpha,angles_min)

            print(len(new_angles),len(new_index))
            return np.array(new_angles,dtype=np.float32), new_index


        else:
            raise NameError('OrdStrat string not recognised: ' + mode)

    # Parse with blocks
    elif blocksize>1:
        # using list comprehension to form the blocks.
        oldindex=np.arange(angles.shape[0])
        index_alpha = np.array([oldindex[i:i+blocksize] for i in range(0,len(oldindex),blocksize)])
        block_alpha = np.array([angles[i:i+blocksize] for i in range(0,angles.shape[0],blocksize)])
        if mode is None or mode == 'ordered':
            return block_alpha, index_alpha

        if mode == 'random'or mode is None:
            new_order = np.arange(len(index_alpha))
            np.random.shuffle(new_order)
            return block_alpha[new_order],index_alpha[[new_order]]
        if mode =='angularDistance':
            # TODO: implement this.
            raise NameError('Angular distance not implemented for blocksize >1 (yet!)')

        else:
            raise NameError('OrdStrat string not recognised: ' + mode)




from __future__ import division
from tigre.algorithms.iterative_recon_alg import IterativeReconAlg
from tigre.algorithms.iterative_recon_alg import decorator
from tigre.utilities.im3Dnorm import im3DNORM
from tigre.algorithms.single_pass_algorithms import FDK
from tigre.utilities.Ax import Ax
import time
import copy
import numpy as np


class ASD_POCS(IterativeReconAlg):
    __doc__ = (" solves the reconstruction problem\n"
               " using the projection data PROJ taken over ALPHA angles, corresponding\n"
               " to the geometry descrived in GEO, using NITER iterations.\n") + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):

        if 'blocksize' not in kwargs:
            kwargs.update(blocksize=1)
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)
        if 'alpha' not in kwargs:
            self.alpha = 0.002
        if 'alpha_red' not in kwargs:
            self.alpha_red = 0.95
        if 'rmax' not in kwargs:
            self.rmax = 0.95
        if 'maxl2err' not in kwargs:
            self.epsilon = im3DNORM(FDK(proj, geo, angles), 2)*0.2
        if "numiter_tv" not in kwargs:
            self.numiter_tv = 20
        if 'regularisation' not in kwargs:
            self.regularisation = 'minimizeTV'
        self.beta = self.lmbda
        self.beta_red = self.lmbda_red

    # Override
    def run_main_iter(self):
        stop_criteria = False
        n_iter = 0
        while not stop_criteria:
            if self.verbose:
                if n_iter == 0:
                    print("POCS Algorithm in progress.")
                    toc = time.clock()
                if n_iter == 1:
                    tic = time.clock()
                    print('Esitmated time until completetion (s): ' + str((self.niter - 1) * (tic - toc)))
            res_prev = copy.deepcopy(self.res)
            n_iter += 1
            getattr(self, self.dataminimizing)()
            g = Ax(self.res, self.geo, self.angles)
            dd = im3DNORM(g-self.proj, 2)
            dp_vec = self.res - res_prev
            dp = im3DNORM(dp_vec, 2)

            if n_iter == 1:
                dtvg = self.alpha * dp

            res_prev = copy.deepcopy(self.res)
            self.res = getattr(self, self.regularisation)(self.res, dtvg)
            dg_vec = self.res - res_prev
            dg = im3DNORM(dg_vec, 2)

            if dg > self.rmax*dp and dd > self.epsilon:
                dtvg = dtvg*self.alpha_red

            self.beta *= self.beta_red
            c = np.dot(dg_vec.reshape(-1,), dp_vec.reshape(-1,))/max(dg*dp, 1e-6)
            if (c < -0.99 and dd <= self.epsilon) or self.beta < 0.005 or n_iter > self.niter:
                if self.verbose:
                    print("\n"
                          "     Stop criteria met: \n"
                          "     c = " + str(c) + "\n"
                          "     beta = " + str(self.beta) + "\n" 
                          "     iter = " + str(n_iter)) + "\n"
                stop_criteria = True


asd_pocs = decorator(ASD_POCS, name='asd_pocs')


class AwASD_POCS(ASD_POCS):
    __doc__ = ASD_POCS.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):

        kwargs.update(dict(regularisation='minimizeAwTV'))

        if 'delta' not in kwargs:
            self.delta = np.array([-0.005], dtype=np.float32)[0]
        ASD_POCS.__init__(self, proj, geo, angles, niter, **kwargs)


awasd_pocs = decorator(AwASD_POCS, name='awasd_pocs')



from __future__ import print_function
from __future__ import division
import numpy as np
import warnings
import scipy.io
def parkerweight(proj,geo,angles,q):
    start=-geo.sDetector[0]/2+geo.dDetector[0]/2
    stop=geo.sDetector[0]/2-geo.dDetector[0]/2
    step=geo.dDetector[0]
    alpha=np.arctan(np.arange(start, stop+step, step)/geo.DSD)
    alpha=-alpha
    delta=abs(alpha[0]-alpha[-1])/2
    totangles=np.cumsum(np.diff(angles))
    totangles=totangles[-1]

    if totangles<2*np.pi:
        warnings.warn('Computing Parker weigths for scanning angle equal or bigger than 2*pi '
              'Consider disabling Parker weigths.')
    if totangles<np.pi+2*delta:
        warnings.warn('Scanning angles smaller than pi+cone_angle. This is limited angle tomgraphy, \n'
                      'there is nosufficient data, thus weigthing for data redundancy is not required.')
    epsilon=max(totangles-(np.pi+2*delta),0)
    data=proj
    # for i in range(proj.shape[0]):
    for i in range(33):
        beta=angles[i]
        w=0.5*(s_function(beta/b_subf(alpha,delta,epsilon,q)-0.5)+s_function((beta-2*delta+2*alpha-epsilon)
                                                                               /b_subf(alpha,delta,epsilon,q)+0.5)
                                                                    -s_function((beta-np.pi+2*alpha)
                                                                                /b_subf(-alpha,delta,epsilon,q)-0.5)
                                                                    -s_function((beta-np.pi-2*delta-epsilon)
                                                                                /b_subf(-alpha,delta,epsilon,q)+0.5)
               )
        proj[i]*=w

    return proj

def s_function(abeta):
    w=np.zeros(abeta.shape)
    w[abeta<=-0.5]=0
    w[abs(abeta)<0.5]=0.5*(1+np.sin(np.pi*abeta[abs(abeta)<0.5]))
    w[abeta>=0.5]=1
    return w
# b_function = B
def b_function(alpha,delta,epsilon):
    return 2*delta-2*alpha+epsilon
# b_subf = b
def b_subf(alpha,delta,epsilon,q):
    return q*b_function(alpha,delta,epsilon)


from __future__ import division, absolute_import, print_function
from .utilities.geometry import geometry
from .utilities.geometry_default import ConeGeometryDefault as geometry_default
from .utilities.Ax import Ax
from .utilities.Atb import Atb
from .utilities.plotproj import plotproj
from .utilities.plotimg import plotimg
from . import algorithms
import tigre.demos



from __future__ import print_function
import matplotlib.pyplot as plt
import numpy as np
def plotproj(projections):
    plt.ion()

    min_val = np.amin(projections)
    max_val = np.amax(projections)
    total_projections = projections.shape[0]
    for i in range(total_projections):
        plt.clf()
        plt.imshow(np.squeeze(projections[i]), cmap=plt.cm.gray, origin='lower', vmin=min_val, vmax=max_val)

        plt.gca().get_xaxis().set_ticks([])
        plt.gca().get_yaxis().set_ticks([])

        plt.xlabel("-> U")
        plt.ylabel("-> V")
        plt.title("Projection angle : " + str(i*360/total_projections))

        plt.colorbar()

        plt.pause(0.001)

def ppslice(projections,slice=None,Dim=2):
    if slice is None:
        slice = projections.shape[2]/2
    min_val = np.amin(projections)
    max_val = np.amax(projections)
    plt.clf()
    if Dim==0:
        plt.imshow(np.squeeze(projections[:,:,slice]), cmap=plt.cm.gray, origin='lower', vmin=min_val, vmax=max_val)
    if Dim==1:
        plt.imshow(np.squeeze(projections[:,slice]), cmap=plt.cm.gray, origin='lower', vmin=min_val, vmax=max_val)
    if Dim==2:
        plt.imshow(np.squeeze(projections[slice]), cmap=plt.cm.gray, origin='lower', vmin=min_val, vmax=max_val)
    plt.colorbar()
    plt.show()




from __future__ import print_function

import numpy as np
import tigre.utilities.geometry_default as geometry
import Test_data.data_loader as data_loader
from tigre.utilities.Ax import Ax
from tigre.utilities.plotproj import ppslice
from tigre.utilities.plotproj import plotproj

geo = geometry.ConeGeometryDefault(high_quality=False)
geo.mode = 'cone'
geo.COR = None
source_img = data_loader.load_head_phantom(number_of_voxels=geo.nVoxel)
angles = np.linspace(0, 2*np.pi, 100, dtype=np.float32)

projsirt = Ax(source_img, geo, angles, 'ray-voxel')
plotproj(projsirt)
#fdk=FDK(projsirt,geo,angles)
# blocksize=input('blocksize:')
niter = 5
# sart=SART(projsirt,geo,angles,niter,init='multigrid',OrderStrategy='angularDistance')
ppslice(projsirt)





from __future__ import division
from matplotlib import pyplot as plt
import numpy as np


class plotimg:
    # NOTE: type help(plotImg) after importing in order to get a readable manual.
    ('\n'
     'plotImg(cube, dim) \n'
     '    plots figure \n'
     'default: progressive in slices following\n'
     '    axis (dim)\n'
     'Parameters \n'
     '---------- \n'
     'cube : Any 3D numpy array \n'
     '\n'
     'dim : ("X","Y","Z","x","y","z"), optional \n'
     '       default is "Z"\n'
     '       NOTE: string arguments!'
     '\n'
     'slice: int, optional\n'
     '     returns page of matrix according to index\n'
     'Examples:\n'
     '---------\n'
     'a=np.ones([3,3,3])\n'
     'plotImg(a)\n'
     '>>>returns plot along dim Z\n'
     'plotImg(a,dim="X")\n'
     '>>>returns plot along dim X\n')

    def __init__(self, cube, dim=None, slice=None):
        self.cube = cube
        self.dim = dim
        self.slice = slice
        self.dimint = None  # keeps track of what dim
        self.dimlist = ['X', 'Y', 'Z', 'x', 'y', 'z', None]  # accepted parameters for dim
        if self.slice is None:
            self.run()
        if self.slice is not None:
            self.slicer()

    def run(self):
        if self.dim not in self.dimlist:
            raise NameError('check inputs for dim, should be string.')

        if self.dim in [None, 'Z', 'z']:
            self.dimint = 2
            self.dimlist = ['X', 'Y', 'Z']
            self.run_plot()
        if self.dim in ['X', 'x']:
            self.dimint = 0
            self.dimlist = ['Y', 'Z', 'X']
            self.run_plot()
        if self.dim in ['Y', 'y']:
            self.dimint = 1
            self.dimlist = ['X', 'Z', 'Y']
            self.run_plot()

    def run_plot(self):
        plt.ion()
        min_val = np.amin(self.cube)
        max_val = np.amax(self.cube)
        dim = self.cube.shape

        for i in range(dim[self.dimint]):
            plt.clf()
            if self.dimint == 2:
                plt.imshow(np.squeeze(self.cube[:, :, i]), cmap=plt.cm.gray, origin='lower', vmin=min_val, vmax=max_val)
            if self.dimint == 1:
                plt.imshow(np.squeeze(self.cube[:, i].transpose()), cmap=plt.cm.gray, origin='lower', vmin=min_val,
                           vmax=max_val)
            if self.dimint == 0:
                plt.imshow(np.squeeze(self.cube[i].transpose()), cmap=plt.cm.gray, origin='lower', vmin=min_val,
                           vmax=max_val)
            plt.gca().get_xaxis().set_ticks([])
            plt.gca().get_yaxis().set_ticks([])
            plt.xlabel(self.dimlist[0])
            plt.ylabel(self.dimlist[1])
            plt.title(self.dimlist[2] + ':' + str(i))

            plt.colorbar()

            plt.pause(0.01)

    def slicer(self):
        # NOTE: Transpose not quite right (still returning image at wrong angle, No labels for axes)
        min_val = np.amin(self.cube)
        max_val = np.amax(self.cube)
        if self.dim in [None, 'Z', 'z']:
            plt.imshow(np.squeeze(self.cube[:,:, self.slice]), cmap=plt.cm.gray, origin='lower', vmin=min_val, vmax=max_val)
        if self.dim in ['Y', 'y']:
            plt.imshow(np.squeeze(self.cube[:, self.slice]), cmap=plt.cm.gray,origin='lower', vmin=min_val, vmax=max_val)
        if self.dim in ['X', 'x']:
            plt.imshow(np.squeeze(self.cube[self.slice]).transpose(), cmap=plt.cm.gray,origin='lower', vmin=min_val, vmax=max_val)
        plt.show()



from __future__ import division
import numpy as np
from tigre.utilities.geometry import Geometry

class ConeGeometryDefault(Geometry):

    def __init__(self, high_quality=True, nVoxel = None):

        Geometry.__init__(self)
        if high_quality:
            # VARIABLE                                          DESCRIPTION                    UNITS
            # -------------------------------------------------------------------------------------
            self.DSD = 1536                                     # Distance Source Detector      (mm)
            self.DSO = 1000                                     # Distance Source Origin        (mm)
            # Detector parameters
            self.nDetector = np.array((512, 512))               # number of pixels              (px)
            self.dDetector = np.array((0.8, 0.8))               # size of each pixel            (mm)
            self.sDetector = self.nDetector * self.dDetector    # total size of the detector    (mm)
            # Image parameters
            self.nVoxel = np.array((512, 512, 512))             # number of voxels              (vx)
            self.sVoxel = np.array((256, 256, 256))             # total size of the image       (mm)
            self.dVoxel = self.sVoxel/self.nVoxel               # size of each voxel            (mm)
            # Offsets
            self.offOrigin = np.array((0, 0, 0))                # Offset of image from origin   (mm)
            self.offDetector = np.array((0, 0))                 # Offset of Detector            (mm)

            # Auxiliary
            self.accuracy = 0.5                                 # Accuracy of FWD proj          (vx/sample)
            # Mode
            self.mode = 'cone'                                  # parallel, cone                ...
            self.filter = None
        else:
            # VARIABLE                                          DESCRIPTION                    UNITS
            # -------------------------------------------------------------------------------------
            self.DSD = 1536                                     # Distance Source Detector      (mm)
            self.DSO = 1000                                     # Distance Source Origin        (mm)
            # Detector parameters
            self.nDetector = np.array((128, 128))                # (V,U) number of pixels        (px)
            self.dDetector = np.array((0.8, 0.8))*4             # size of each pixel            (mm)
            self.sDetector = self.nDetector * self.dDetector    # total size of the detector    (mm)
            # Image parameters
            self.nVoxel = np.array((64,  64, 64))               # number of voxels              (vx)
            self.sVoxel = np.array((256, 256, 256))             # total size of the image       (mm)
            self.dVoxel = self.sVoxel / self.nVoxel             # size of each voxel            (mm)
            # Offsets
            self.offOrigin = np.array((0, 0, 0))                # Offset of image from origin   (mm)
            self.offDetector = np.array((0, 0))                 # Offset of Detector            (mm)
            self.rotDetector = np.array((0,0,0))
            # Auxiliary
            self.accuracy = 0.5                                 # Accuracy of FWD proj          (vx/sample)
            # Mode
            self.mode='cone'                                    # parallel, cone                ...
            self.filter=None
        if nVoxel is not None:
            # VARIABLE                                          DESCRIPTION                    UNITS
            # -------------------------------------------------------------------------------------
            self.DSD = 1536                                     # Distance Source Detector      (mm)
            self.DSO = 1000                                     # Distance Source Origin        (mm)
                                                                # Detector parameters
            self.nDetector = np.array((128, 128))               # (V,U) number of pixels        (px)
            self.dDetector = np.array((0.8, 0.8)) * 4           # size of each pixel            (mm)
            self.sDetector = self.nDetector * self.dDetector        # total size of the detector    (mm)
                                                                # Image parameters
            self.nVoxel = np.array((nVoxel))                # number of voxels              (vx)
            self.sVoxel = np.array((256, 256, 256))             # total size of the image       (mm)
            self.dVoxel = self.sVoxel / self.nVoxel             # size of each voxel            (mm)
            # Offsets
            self.offOrigin = np.array((0, 0, 0))                # Offset of image from origin   (mm)
            self.offDetector = np.array((0, 0))                 # Offset of Detector            (mm)
            self.rotDetector = np.array((0, 0, 0))
            # Auxiliary
            self.accuracy = 0.5                                 # Accuracy of FWD proj          (vx/sample)
            # Mode
            self.mode = 'cone'                                  # parallel, cone



#   This file is part of the TIGRE Toolbox

#   Copyright (c) 2015, University of Bath and
#                       CERN-European Organization for Nuclear Research
#                       All rights reserved.

#   License:            Open Source under BSD.
#                       See the full license at
#                       https://github.com/CERN/TIGRE/license.txt

#   Contact:            tigre.toolbox@gmail.com
#   Codes:              https://github.com/CERN/TIGRE/
# --------------------------------------------------------------------------
#   Coded by:          MATLAB (original code): Ander Biguri
#                      PYTHON : Sam Loescher, Reuben Lindroos

cimport numpy as np 
import numpy as np
from tigre.utilities.errors import TigreCudaCallError
np.import_array()

from libc.stdlib cimport malloc, free 

cdef extern from "numpy/arrayobject.h":
    void PyArray_ENABLEFLAGS(np.ndarray arr, int flags)
    void PyArray_CLEARFLAGS(np.ndarray arr, int flags)

cdef extern from "tvdenoising.hpp":
    cdef void tvdenoising(float* src, float* dst, float lamda, float* spacing, long* image_size, int maxiter)


def cuda_raise_errors(error_code):
    if error_code:
        raise TigreCudaCallError('tvdenoising:',error_code)


def tvdenoise(np.ndarray[np.float32_t, ndim=3] src, int maxiter = 100, float lamda = 15.0):

    
    cdef np.npy_intp size_img[3]
    size_img[0]= <np.npy_intp> src.shape[0]
    size_img[1]= <np.npy_intp> src.shape[1]
    size_img[2]= <np.npy_intp> src.shape[2]

    cdef float* c_imgout = <float*> malloc(size_img[0] *size_img[1] *size_img[2]* sizeof(float))

    cdef float spacing[3]
    spacing[0]=<float> 1
    spacing[1]=<float> 1
    spacing[2]=<float> 1

    cdef long imgsize[3]
    imgsize[0] = <long> size_img[0]
    imgsize[1] = <long> size_img[1]
    imgsize[2] = <long> size_img[2]

    cdef float* c_src = <float*> src.data
    cdef np.npy_intp c_maxiter = <np.npy_intp> maxiter
    cuda_raise_errors(tvdenoising(c_src, c_imgout, lamda, spacing, imgsize, c_maxiter))

    imgout = np.PyArray_SimpleNewFromData(3, size_img, np.NPY_FLOAT32, c_imgout)
    PyArray_ENABLEFLAGS(imgout, np.NPY_OWNDATA)

    return imgout


from __future__ import division
from __future__ import print_function
import numpy as np
import numpy.matlib as matlib
import inspect
import tigre
import math


class Geometry(object):

    def __init__(self):
        self.mode = None
        self.n_proj = None
        self.angles = None
        self.filter = None

    def check_geo(self, angles, verbose=False):
        if angles.ndim == 1:
            self.n_proj = angles.shape[0]
            zeros_array = np.zeros((self.n_proj, 1), dtype=np.float32)
            self.angles = np.hstack((angles.reshape(self.n_proj, 1), zeros_array, zeros_array))

        elif angles.ndim == 2:
            if angles.shape[1] != 3:
                raise BufferError("Expected angles of dimensions (n, 3), got: " + str(angles.shape))
            self.n_proj = angles.shape[0]
            angles = angles.copy()
            setattr(self, 'angles', angles)
        else:
            raise BufferError("Unexpected angles shape: " + str(angles.shape))
        if self.mode is None:
            setattr(self, 'mode', 'cone')

        manditory_attribs = ['nVoxel', 'sVoxel', 'dVoxel',
                             'nDetector', 'sDetector', 'dDetector',
                             'DSO', 'DSD']
        included_attribs_indx = [hasattr(self, attrib) for attrib in manditory_attribs]
        if not all(included_attribs_indx):
            raise AttributeError('following manditory fields '
                                 'missing from geometry:' + str([attrib for attrib in manditory_attribs
                                                                 if not hasattr(self, attrib)])
                                 )
        optional_attribs = ['offOrigin', 'offDetector', 'rotDetector', 'COR',
                            'mode', 'accuracy']

        # image data
        if not self.nVoxel.shape == (3,): raise AttributeError('geo.nVoxel.shape should be (3, )')
        if not self.sVoxel.shape == (3,): raise AttributeError('geo.sVoxel.shape should be (3, )')
        if not self.dVoxel.shape == (3,): raise AttributeError('geo.dVoxel.shape should be (3, )')
        if not sum(abs(self.dVoxel * self.nVoxel - self.sVoxel)) < 1e-6: 'nVoxel*dVoxel is not equal to sVoxel. ' \
                                                                         'Check fields.'

        # Detector Data
        if not self.nDetector.shape == (2,): raise AttributeError('geo.nDecetor.shape should be (2, )')
        if not self.sDetector.shape == (2,): raise AttributeError('geo.sDetector.shape should be (2, )')
        if not self.dDetector.shape == (2,): raise AttributeError('geo.dDetector.shape should be (2, )')
        if not sum(abs(self.dDetector * self.nDetector - self.sDetector)) < 1e-6: raise AttributeError(
            'nDetector*dDetecor is not equal to sDetector. Check fields.')

        for attrib in ['DSD', 'DSO']:
            self.__check_and_repmat__(attrib, angles)

        if hasattr(self, 'offOrigin'):
            self.__check_and_repmat__('offOrigin', angles)
        else:
            self.offOrigin = np.array([0, 0, 0])
            self.__check_and_repmat__('offOrigin', angles)

        if hasattr(self, 'offDetector'):
            self.__check_and_repmat__('offDetector', angles)
        else:
            self.offDetector = np.array([0, 0])
            self.offDetector = np.zeros((angles.shape[0], 2))

        if hasattr(self, 'rotDetector'):
            self.__check_and_repmat__('rotDetector', angles)
        else:
            self.rotDetector = np.array([0, 0, 0])
            self.__check_and_repmat__('rotDetector', angles)

        if hasattr(self, 'COR'):
            self.__check_and_repmat__('COR', angles)
        else:
            self.COR = np.zeros(angles.shape[0])
        # IMPORTANT: cast all numbers to float32
        if verbose:
            self._verbose_output()

    def checknans(self):
        for attrib in self.__dict__:
            if str(getattr(self, attrib)) == 'nan':
                raise ValueError('nan found for Geometry abbtribute:' + attrib)
            elif type(getattr(self, attrib)) == np.ndarray:
                if np.isnan(getattr(self, attrib)).all():
                    raise ValueError('Nan found in Geometry abbtribute:' + attrib)

    def cast_to_single(self):
        """
        Casts all number values in current instance to
        single prevision floating point types.
        :return: None
        """
        for attrib in self.__dict__:
            if getattr(self, attrib) is not None:
                try:
                    setattr(self, attrib, np.float32(getattr(self, attrib)))
                except ValueError:
                    pass

    def __check_and_repmat__(self, attrib, angles):
        """
        Checks whether the attribute is a single value and repeats it into an array if it is
        :rtype: None
        :param attrib: string
        :param angles: np.ndarray
        """
        old_attrib = getattr(self, attrib)

        if type(old_attrib) in [float, int, np.float32, np.float64]:
            new_attrib = matlib.repmat(old_attrib, 1, angles.shape[0])[0]
            setattr(self, attrib, new_attrib)

        elif type(old_attrib) == np.ndarray:
            if old_attrib.ndim == 1:
                if old_attrib.shape in [(3,), (2,), (1,)]:
                    new_attrib = matlib.repmat(old_attrib, angles.shape[0], 1)
                    setattr(self, attrib, new_attrib)
                elif old_attrib.shape == (angles.shape[0],):
                    pass
            else:
                if old_attrib.shape == (angles.shape[0], old_attrib.shape[1]):
                    pass
                else:
                    raise AttributeError(attrib + " with shape: " + str(old_attrib.shape) +
                                         " not compatible with shapes: " + str([(angles.shape[0],),
                                                                                (angles.shape[0], old_attrib.shape[1]),
                                                                                (3,), (2,), (1,)]))

        else:
            raise TypeError(
                "Data type not understood for: geo." + attrib + " with type = " + str(type(getattr(self, attrib))))

    def _verbose_output(self):
        for obj in inspect.getmembers(self):
            if obj[0][0] == '_':
                pass
            elif obj[0] == 'check_geo':
                pass
            elif type(obj[1]) == np.ndarray:
                print(self.mode + ': ' + str((obj[0], obj[1].shape)))
            else:
                print(self.mode + ': ' + str(obj))

    def convert_contig_mode(self):
        dim_attribs = ['nVoxel', 'sVoxel', 'dVoxel',
                       'nDetector', 'sDetector', 'dDetector']
        for attrib in dim_attribs:
            setattr(self, attrib, getattr(self, attrib)[::-1].copy())

    def __str__(self):
        parameters = []
        parameters.append("TIGRE parameters")
        parameters.append("-----")
        parameters.append("Geometry parameters")
        parameters.append("Distance from source to detector (DSD) = " + str(self.DSD) + " mm")
        parameters.append("Distance from source to origin (DSO)= " + str(self.DSO) + " mm")
        parameters.append("-----")
        parameters.append("Detector parameters")
        parameters.append("Number of pixels (nDetector) = " + str(self.nDetector))
        parameters.append("Size of each pixel (dDetector) = " + str(self.dDetector) + " mm")
        parameters.append("Total size of the detector (sDetector) = " + str(self.sDetector) + " mm")
        parameters.append("-----")
        parameters.append("Image parameters")
        parameters.append("Number of voxels (nVoxel) = " + str(self.nVoxel))
        parameters.append("Total size of the image (sVoxel) = " + str(self.sVoxel) + " mm")
        parameters.append("Size of each voxel (dVoxel) = " + str(self.dVoxel) + " mm")

        parameters.append("-----")
        if hasattr(self, 'offOrigin') and hasattr(self, 'offDetector'):
            parameters.append("Offset correction parameters")
            if hasattr(self, 'offOrigin'):
                parameters.append("Offset of image from origin (offOrigin) = " + str(self.offOrigin) + " mm")
            if hasattr(self, 'offDetector'):
                parameters.append("Offset of detector (offDetector) = " + str(self.offDetector) + " mm")

        parameters.append("-----")
        parameters.append("Auxillary parameters")
        parameters.append("Samples per pixel of forward projection (accuracy) = " + str(self.accuracy))

        if hasattr(self, 'rotDetector'):
            parameters.append("-----")
            parameters.append("Rotation of the Detector (rotDetector) = " + str(self.rotDetector) + " rad")

        if hasattr(self, 'COR'):
            parameters.append("-----")
            parameters.append("Centre of rotation correction (COR) = " + str(self.COR) + " mm")

        return '\n'.join(parameters)

    def __cmp__(self, other):
        resultofnumpiesanallyretentiveattemptatbeingphilosophical = []
        for attrib in self.__dict__:
            result = (getattr(self, attrib) == getattr(other, attrib))
            try:
                resultofnumpiesanallyretentiveattemptatbeingphilosophical.append(result.all())
            except Exception:
                try:
                    resultofnumpiesanallyretentiveattemptatbeingphilosophical.extend(result)
                except Exception:
                    resultofnumpiesanallyretentiveattemptatbeingphilosophical.append(result)

        # why is this boolean reversed when returned?
        # because for some reason its reversed when i return it from this function. Who knows.

        return not all(resultofnumpiesanallyretentiveattemptatbeingphilosophical)


class ParallelGeo(Geometry):
    def __init__(self, nVoxel):
        if nVoxel is None:
            raise ValueError('nVoxel needs to be given for initialisation of parallel beam')
        Geometry.__init__(self)
        self.mode = 'parallel'

        self.nVoxel = nVoxel
        self.dVoxel = np.array([1, 1, 1])
        self.sVoxel = self.nVoxel

        self.DSO = np.float32(self.nVoxel[0])
        self.DSD = np.float32(self.nVoxel[0] * 2)

        self.dDetector = np.array([1, 1])
        self.nDetector = np.zeros((2,),dtype = np.int64)
        self.nDetector[0] = self.nVoxel[0]
        self.nDetector[1] = max(self.nVoxel[1],self.nVoxel[2])
        #self.nDetector = self.nVoxel[:2]
        self.sDetector = self.nDetector

        self.accuracy = 0.5

        self.offOrigin = np.array([0, 0, 0]);
        self.offDetector = np.array([0, 0]);

        self.rotDetector = np.array([0, 0, 0])


def geometry(mode='cone', nVoxel=None, default_geo=False, high_quality=True):
    if mode == 'cone':
        if default_geo:
            return tigre.geometry_default(high_quality, nVoxel)
        else:
            return Geometry()
    if mode == 'parallel':
        return ParallelGeo(nVoxel)
    else:
        raise ValueError('mode: ' + mode + ' not recognised.')



from _Atb import _Atb_ext
import numpy as np
import copy
def Atb(projections, geo, angles,  krylov="matched"):

    if projections.dtype != np.float32:
        raise TypeError("Input data should be float32, not " + str(projections.dtype))
    if not np.isreal(projections).all():
        raise ValueError("Complex types not compatible for back projection.")
    geox = copy.deepcopy(geo)
    geox.check_geo(angles)
    """
       Here we cast all values in geo to single point precision float. This way we 
       know what behaviour to expect from pytigre to Cuda and can change 
       single parameters accordingly. 
    """
    geox.cast_to_single()
    geox.checknans()

    if projections.shape != (geox.angles.shape[0],geox.nDetector[0],geox.nDetector[1]):
        raise ValueError("Expected data shape: " + str((geox.angles.shape[0], geox.nDetector[0],geox.nDetector[1])) +
                         " not compatible with: " + str(projections.shape))

    return _Atb_ext(projections, geox, geox.angles, krylov,geox.mode)


import os
def test():
    """
    Function only outputs to terminal but gives good coverage of
    the functions implemented and works well as a unit test for
    imports.

    Usage
    --------
    import tigre
    tigre.demos.test()

    :return: None
    """
    dirname = os.path.dirname(__file__)
    for filename in os.listdir(dirname):
        if filename.startswith('d0'):
            """
            cmd = ['jupyter nbconvert --to notebook --execute --ExecutePreprocessor.timeout=60 ' +
                      os.path.join(dirname,filename)]
            print(cmd)
            output=subprocess.Popen(cmd, stdout=subprocess.PIPE,shell=True).communicate()[0]
            nb = nbformat.read(output, nbformat.current_nbformat)
            """
            os.system('jupyter nbconvert --to notebook --execute --ExecutePreprocessor.timeout=60 ' +
                      os.path.join(dirname,filename))

    cleanup()

def cleanup():
    dirname = os.path.dirname(__file__)
    for filename in os.listdir(dirname):
        if filename.endswith('nbconvert.ipynb'):
            os.system('rm ' + os.path.join(dirname,filename))


from _Ax import _Ax_ext
import numpy as np
import copy

def Ax(img, geo, angles,  krylov="ray-voxel"):

    if img.dtype != np.float32:
        raise TypeError("Input data should be float32, not "+ str(img.dtype))
    if not np.isreal(img).all():
        raise ValueError("Complex types not compatible for projection.")
    geox = copy.deepcopy(geo)
    geox.check_geo(angles)
    """
    Here we cast all values in geo to single point precision float. This way we 
    know what behaviour to expect from pytigre to Cuda and can change 
    single parameters accordingly. 
    """
    geox.cast_to_single()
    geox.checknans()

    if all(img.shape != geox.nVoxel):
        raise ValueError("Input data should be of shape geo.nVoxel: "+ str(geox.nVoxel) +
                         " not:" + str(img.shape))

    return _Ax_ext(img, geox, geox.angles, krylov, geox.mode)



import numpy as np

def im3DNORM(img,normind,varargin=None):
    '''
    % IMAGE3DNORM computes the desired image norm
    %   IMAGE3DNORM(IMG,NORMIND) computes the norm if image IMG using the norm
    %   defined in NORMING
    %
    %   IMG         A 3D image
    %   NORMIND     {non-zero int, inf, -inf, 'fro', 'nuc'}
    %               'TV': TV norm
    %
    %
    %--------------------------------------------------------------------------
    %--------------------------------------------------------------------------
    % This file is part of the TIGRE Toolbox
    %
    % Copyright (c) 2015, University of Bath and
    %                     CERN-European Organization for Nuclear Research
    %                     All rights reserved.
    %
    % License:            Open Source under BSD.
    %                     See the full license at
    %                     https://github.com/CERN/TIGRE/license.txt
    %
    % Contact:            tigre.toolbox@gmail.com
    % Codes:              https://github.com/CERN/TIGRE/
    % Coded by:           Ander Biguri
    %--------------------------------------------------------------------------
    '''

    if normind is [np.inf,-np.inf,'fro','nuc']:
        return np.linalg.norm(img.ravel(),normind)
    if type(normind) is int:
        return np.linalg.norm(img.ravel(),normind)
    if normind is 'TV':
        raise Exception('TV not yet implemented')





import os
import sys
# TODO: this is quite nasty; it would be nice to reorganise file structure later so top level folder is always in path
currDir = os.path.dirname(os.path.realpath(__file__))
rootDir = os.path.abspath(os.path.join(currDir, '..'))
if rootDir not in sys.path:  # add parent dir to paths
    sys.path.append(rootDir)

import numpy as np
import tigre.geometry_default as geometry
import Test_data.data_loader as data_loader
import scipy.io
from tigre.utilities.plotproj import plotproj

from _Ax import Ax

TIGRE_parameters = geometry(high_quality=False)

head = data_loader.load_head_phantom(number_of_voxels=TIGRE_parameters.nVoxel)

angles = np.linspace(0, 2*np.pi, 20, dtype=np.float32)
# angles = np.linspace(0, 0, 1, dtype=np.float32)
# What is expected to happen here is the memory allocated for the head image is given to the C functions to operate on
# Whilst it will return newly allocated PYTHON(!) memory for each projection; expected to be allocated from C and parsed
# back into python

# I believe it should be ok to just parse geometry each time until we find a way of passing around a C array in python

projections = Ax(head, TIGRE_parameters, angles, 'interpolated')

plotproj(projections)

# m = {'projections': projections}
# scipy.io.savemat('projections_python', m)



from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import os
import sys
import numpy as np
import copy
from tigre.utilities.Atb import Atb
from tigre.utilities.filtering import filtering

# TODO: this is quite nasty; it would be nice to reorganise file structure later so top level folder is always in path
currDir = os.path.dirname(os.path.realpath(__file__))
rootDir = os.path.abspath(os.path.join(currDir, '..'))
if rootDir not in sys.path:  # add parent dir to paths
    sys.path.append(rootDir)


def FDK(proj, geo, angles, filter='ram_lak', verbose=False, **kwargs):
    """
    solves CT image reconstruction.

    :param proj: np.array(dtype=float32),
     Data input in the form of 3d

    :param geo: tigre.utilities.geometry.Geometry
     Geometry of detector and image (see examples/Demo code)

    :param angles: np.array(dtype=float32)
     Angles of projection, shape = (nangles,3) or (nangles,)

    :param filter: str
     Type of filter used for backprojection
     opts: "shep_logan"
           "cosine"
           "hamming"
           "hann"

    :param verbose: bool
     Feedback print statements for algorithm progress

    :param kwargs: dict
     keyword arguments

    :return: np.array(dtype=float32)

    Usage:
    -------
    >>> import tigre
    >>> import tigre.algorithms as algs
    >>> import numpy
    >>> from tigre.demos.Test_data import data_loader
    >>> geo = tigre.geometry(mode='cone',default_geo=True,
    >>>                         nVoxel=np.array([64,64,64]))
    >>> angles = np.linspace(0,2*np.pi,100)
    >>> src_img = data_loader.load_head_phantom(geo.nVoxel)
    >>> proj = tigre.Ax(src_img,geo,angles)
    >>> output = algs.FDK(proj,geo,angles)

    tigre.demos.run() to launch ipython notebook file with examples.


    --------------------------------------------------------------------
    This file is part of the TIGRE Toolbox

    Copyright (c) 2015, University of Bath and
                        CERN-European Organization for Nuclear Research
                        All rights reserved.

    License:            Open Source under BSD.
                        See the full license at
                        https://github.com/CERN/TIGRE/license.txt

    Contact:            tigre.toolbox@gmail.com
    Codes:              https://github.com/CERN/TIGRE/
  ----------------------------------------------------------------------
    Coded by:          MATLAB (original code): Ander Biguri
                       PYTHON : Reuben Lindroos


    """

    if 'niter' in kwargs:
        kwargs.pop('niter')

    geo = copy.deepcopy(geo)
    geo.check_geo(angles)
    geo.checknans()
    if filter is not None:
        geo.filter = filter
    # Weight
    proj_filt = np.zeros(proj.shape, dtype=np.float32)
    for ii in range(angles.shape[0]):
        xv = np.arange((-geo.nDetector[1] / 2) + 0.5, 1 + (geo.nDetector[1] / 2) - 0.5) * geo.dDetector[1]
        yv = np.arange((-geo.nDetector[0] / 2) + 0.5, 1 + (geo.nDetector[0] / 2) - 0.5) * geo.dDetector[0]
        (yy, xx) = np.meshgrid(xv, yv)

        w = geo.DSD[0] / np.sqrt((geo.DSD[0] ** 2 + xx ** 2 + yy ** 2))
        proj_filt[ii] = copy.deepcopy(proj[ii]) * w

    proj_filt = filtering(proj_filt, geo, angles, parker=False, verbose=verbose)
    # m = {
    #     'py_projfilt': proj_filt,
    #
    # }
    # scipy.io.savemat('Tests/Filter_data', m)
    res = Atb(proj_filt, geo, geo.angles, 'FDK')
    # res = 0
    # res = Atb(proj,geo,angles,'FDK')
    return res


fdk = FDK


def fbp(proj, geo, angles, filter=None, verbose=False, **kwargs):
    __doc__ = FDK.__doc__
    if geo.mode != 'parallel':
        raise ValueError("Only use FBP for parallel beam. Check geo.mode.")
    geox = copy.deepcopy(geo)
    geox.check_geo(angles)
    proj_filt = filtering(copy.deepcopy(proj), geox, angles, parker=False, verbose=verbose)
    res = Atb(proj_filt, geo, angles) * geo.DSO / geo.DSD

    return res



from __future__ import print_function
import tigre
import numpy as np
from testforlog import testandlog
import sys
import traceback
nVoxel = np.array([64, 64, 64])


def configuration1(alg):
    try:

        geo = tigre.geometry(mode='cone', nVoxel=nVoxel, default_geo=True)
        niter = 10
        nangles = 100
        angles_1 = np.linspace(0, 2 * np.pi, nangles, dtype=np.float32)
        angles_2 = np.zeros((nangles), dtype=np.float32)
        angles = np.vstack((angles_1, angles_2, angles_2)).T
        testandlog([alg], geo, angles, niter, saveresult=True, subdirectoryname='configuration1', sup_kw_warning=True)


    except Exception as e:
        print(traceback.format_exc())
        raise SystemExit()


def configuration2(alg):
    try:
        geo = tigre.geometry(mode='parallel', nVoxel=nVoxel, default_geo=True)
        niter = 10
        nangles = 100
        angles_1 = np.linspace(0, 2 * np.pi, nangles, dtype=np.float32)
        angles_2 = np.zeros((nangles), dtype=np.float32)
        angles = np.vstack((angles_1, angles_2, angles_2)).T
        testandlog([alg], geo, angles, niter, saveresult=True, subdirectoryname='configuration2', sup_kw_warning=True)
    except Exception as e:
        print(traceback.format_exc())
        raise SystemExit()


def configuration3(alg):
    try:
        geo = tigre.geometry(mode='cone', nVoxel=np.array([32, 64, 128]), default_geo=True)
        niter = 10
        nangles = 100
        angles_1 = np.linspace(0, 2 * np.pi, nangles, dtype=np.float32)
        angles_2 = np.zeros((nangles), dtype=np.float32)
        angles = np.vstack((angles_1, angles_2, angles_2)).T
        testandlog([alg], geo, angles, niter, saveresult=True, subdirectoryname='configuration3', sup_kw_warning=True)
    except Exception as e:
        print(traceback.format_exc())
        raise SystemExit()


def configuration4(alg):
    try:
        geo = tigre.geometry(mode='parallel', nVoxel=np.array([32, 64, 128]), default_geo=True)
        niter = 10
        nangles = 100
        angles_1 = np.linspace(0, 2 * np.pi, nangles, dtype=np.float32)
        angles_2 = np.zeros((nangles), dtype=np.float32)
        angles = np.vstack((angles_1, angles_2, angles_2)).T
        testandlog([alg], geo, angles, niter, saveresult=True, subdirectoryname='configuration4', sup_kw_warning=True)
    except Exception as e:
        print(traceback.format_exc())
        raise SystemExit()


def configuration5(alg):
    try:
        geo = tigre.geometry(mode='cone', nVoxel=np.array([32, 64, 128]), default_geo=True)
        niter = 10
        nangles = 100
        angles_1 = np.linspace(0, 2 * np.pi, nangles, dtype=np.float32)
        angles_2 = np.ones((nangles), dtype=np.float32) * np.pi
        angles = np.vstack((angles_1, angles_2, angles_2)).T
        testandlog([alg], geo, angles, niter, saveresult=True, subdirectoryname='configuration5', sup_kw_warning=True)
    except Exception as e:
        print(traceback.format_exc())
        raise SystemExit()


if __name__ == '__main__':
    method_name = sys.argv[1]
    parameter_name = sys.argv[2]
    getattr(sys.modules[__name__], method_name)(parameter_name)



from __future__ import division, absolute_import, print_function
from .art_family_algorithms import sart
from .art_family_algorithms import sirt
from .art_family_algorithms import ossart
from .iterative_recon_alg import iterativereconalg
from .single_pass_algorithms import FDK
from .pocs_algorithms import asd_pocs
from .pocs_algorithms import awasd_pocs
from .single_pass_algorithms import fbp
from .single_pass_algorithms import fdk
from .conjugate_gradient_algorithms import cgls

__all__ = ['sart',
           'sirt',
           'ossart',
           'iterativereconalg',
           'FDK',
           'asd_pocs',
           'awasd_pocs',
           'fbp',
           'cgls']



from tigre.algorithms.iterative_recon_alg import IterativeReconAlg
from tigre.algorithms.iterative_recon_alg import decorator


class SART(IterativeReconAlg):
    __doc__ = ("SART_CBCT solves Cone Beam CT image reconstruction using Oriented Subsets\n"
               "Simultaneous Algebraic Reconstruction Techique algorithm\n"
               "SART(PROJ,GEO,ALPHA,NITER) solves the reconstruction problem\n"
               "using the projection data PROJ taken over ALPHA angles, corresponding\n"
               "to the geometry described in GEO, using NITER iterations. \n") + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):
        kwargs.update(dict(blocksize=1))
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)


sart = decorator(SART, name='sart')


class SIRT(IterativeReconAlg):
    __doc__ = ("SART_CBCT solves Cone Beam CT image reconstruction using Oriented Subsets\n"
               "Simultaneous Algebraic Reconxtruction Techique algorithm\n"

               "SIRT(PROJ,GEO,ALPHA,NITER) solves the reconstruction problem\n"
               "using the projection data PROJ taken over ALPHA angles, corresponding\n"
               "to the geometry descrived in GEO, using NITER iterations.\n") + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):
        kwargs.update(dict(blocksize=angles.shape[0]))
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)


sirt = decorator(SIRT, name='sirt')


class OS_SART(IterativeReconAlg):
    __doc__ = ("SART_CBCT solves Cone Beam CT image reconstruction using Oriented Subsets\n"
               "Simultaneous Algebraic Reconxtruction Techique algorithm\n"

               "OS_SART(PROJ,GEO,ALPHA,NITER,BLOCKSIZE=20) solves the reconstruction problem\n"
               "using the projection data PROJ taken over ALPHA angles, corresponding\n"
               "to the geometry descrived in GEO, using NITER iterations.\n") + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)


ossart = decorator(OS_SART, name='ossart')



#   This file is part of the TIGRE Toolbox

#   Copyright (c) 2015, University of Bath and
#                       CERN-European Organization for Nuclear Research
#                       All rights reserved.

#   License:            Open Source under BSD.
#                       See the full license at
#                       https://github.com/CERN/TIGRE/license.txt

#   Contact:            tigre.toolbox@gmail.com
#   Codes:              https://github.com/CERN/TIGRE/
# --------------------------------------------------------------------------
#   Coded by:          MATLAB (original code): Ander Biguri
#                      PYTHON : Sam Loescher, Reuben Lindroos

cimport numpy as np 
import numpy as np

np.import_array()
from libc.stdlib cimport malloc, free 

cdef extern from "numpy/arrayobject.h":
    void PyArray_ENABLEFLAGS(np.ndarray arr, int flags)
    void PyArray_CLEARFLAGS(np.ndarray arr, int flags)

cdef extern from "POCS_TV2.hpp":
    cdef int aw_pocs_tv(float* img, float* dst, float alpha, long* image_size, int maxiter, float delta)
def cuda_raise_errors(error_code):
    if error_code:
        raise ValueError('TIGRE:Call to aw_pocs_tv failed')

def AwminTV(np.ndarray[np.float32_t, ndim=3] src,float alpha = 15.0,int maxiter = 100, float delta=-0.005):
    cdef np.npy_intp size_img[3]
    size_img[0]= <np.npy_intp> src.shape[0]
    size_img[1]= <np.npy_intp> src.shape[1]
    size_img[2]= <np.npy_intp> src.shape[2]

    cdef float* c_imgout = <float*> malloc(size_img[0] *size_img[1] *size_img[2]* sizeof(float))

    cdef long imgsize[3]
    imgsize[0] = <long> size_img[0]
    imgsize[1] = <long> size_img[1]
    imgsize[2] = <long> size_img[2]

    cdef float* c_src = <float*> src.data
    cdef np.npy_intp c_maxiter = <np.npy_intp> maxiter
    cuda_raise_errors(aw_pocs_tv(c_src, c_imgout, alpha, imgsize, c_maxiter, delta))
    imgout = np.PyArray_SimpleNewFromData(3, size_img, np.NPY_FLOAT32, c_imgout)
    PyArray_ENABLEFLAGS(imgout, np.NPY_OWNDATA)

    return imgout


from __future__ import print_function
import os
import sys
import tigre

import numpy as np
from matplotlib import pyplot as plt
from tigre.demos.Test_data import data_loader
from tigre.utilities.Ax import Ax
import tigre.algorithms as algs
from tigre.algorithms.iterative_recon_alg import IterativeReconAlg

# ----------------PROFILINGIMPORT-------------------

from line_profiler import LineProfiler

# ---------------GEOMETRY---------------------------

geo = tigre.geometry(mode='parallel', nVoxel=np.array([64, 64, 64]))
source_img = data_loader.load_head_phantom(number_of_voxels=geo.nVoxel)

# ---------------------ANGLES-------------------------

angles_1 = np.linspace(0, 2 * np.pi, 100, dtype=np.float32)
angles_2 = np.ones((100), dtype=np.float32) * np.array(np.pi / 4, dtype=np.float32)
angles_3 = np.zeros((100), dtype=np.float32)
angles = np.vstack((angles_1, angles_3, angles_3)).T

# --------------------PROJECTION----------------------

proj = Ax(source_img, geo, angles)


# ---------------------RECONSTRUCTION------------------

class lineprofileroveride(IterativeReconAlg):
    def __init__(self, proj, geo, angles, niter, **kwargs):
        lp = LineProfiler()
        lp_wrapper = lp(super(lineprofileroveride, self).__init__)
        lp_wrapper(proj, geo, angles, niter, **kwargs)
        lp.print_stats()


alg = lineprofileroveride(proj, geo, angles, 10, **dict(blocksize=10))

# sart = IterativeReconAlg(proj,geo,angles,niter=10, **dict(blocksize =20))
# lp_wrapper = lp(getattr(sart,sart.dataminimizing))
# setattr(sart, sart.dataminimizing, lp_wrapper)
# sart.run_main_iter()
# lp.print_stats()

from tigre.algorithms.pocs_algorithms import ASD_POCS

lp = LineProfiler()
asd_pocs = ASD_POCS(proj, geo, angles, 2, **dict(blocksize=20))
lp_wrapper_data_minimizing = lp(getattr(asd_pocs, asd_pocs.dataminimizing))
setattr(asd_pocs, asd_pocs.dataminimizing, lp_wrapper_data_minimizing)
lp_wrapper_main_iter = lp(asd_pocs.run_main_iter)
lp_wrapper_main_iter()
lp.print_stats()

# -------------Line profile GGLS------------------

from tigre.algorithms.conjugate_gradient_algorithms import CGLS
from line_profiler import LineProfiler

cglsAlg = CGLS(proj, geo, angles, niter=20)
lp = LineProfiler()
lp_run_main_iter = lp(cglsAlg.run_main_iter)
lp_run_main_iter()
lp.print_stats()



from __future__ import division

import numpy as np
from tigre.utilities.Ax import Ax
from tigre.utilities.Atb import Atb
from tigre.utilities.order_subsets import order_subsets
from tigre.utilities.init_multigrid import init_multigrid
from tigre.utilities.Measure_Quality import Measure_Quality as MQ
from tigre.utilities.im3Dnorm import im3DNORM
from tigre.algorithms.single_pass_algorithms import FDK
from _minTV import minTV
from _AwminTV import AwminTV
import time
import copy
"""
This module is where the umbrella class IterativeReconAlg is located
which is the umbrella class to all the other algorithms apart from 
the single pass type algorithms. 
"""

# coding: utf8


class IterativeReconAlg(object):
    """
    Parameters
    ----------
    :param proj: (np.ndarray, dtype=np.float32)
    Input data, shape = (geo.nDector, nangles)

    :param geo: (tigre.geometry)
    Geometry of detector and image (see examples/Demo code)

    :param angles: (np.ndarray , dtype=np.float32)
    angles of projection, shape = (nangles,3)

    :param niter: (int)
    number of iterations for reconstruction algorithm

    :param kwargs: (dict)
    optional parameters

    Keyword Arguments
    -----------------
    :keyword blocksize: (int)
        number of angles to be included in each iteration
        of proj and backproj for OS_SART
    :keyword lmbda: (np.float64)
        Sets the value of the hyperparameter.

    :keyword lmbda_red: (np.float64)
        Reduction of lambda every iteration
        lambda=lambdared*lambda. Default is 0.99

    :keyword init: (str)
        Describes different initialization techniques.
              "none"     : Initializes the image to zeros (default)
              "FDK"      : intializes image to FDK reconstrucition
              "multigrid": Initializes image by solving the problem in
                           small scale and increasing it when relative
                           convergence is reached.
              "image"    : Initialization using a user specified
                           image. Not recommended unless you really
                           know what you are doing.

    :keyword InitImg: (np.ndarray)
        Not yet implemented. Image for the "image" initialization.

    :keyword verbose:  (Boolean)
        Feedback print statements for algorithm progress
        default=True

    :keyword Quameasopts: (list)
        Asks the algorithm for a set of quality measurement
        parameters. Input should contain a list or tuple of strings of
        quality measurement names. Examples:
            RMSE, CC, UQI, MSSIM

    :keyword OrderStrategy : (str)
        Chooses the subset ordering strategy. Options are:
                 "ordered"        : uses them in the input order, but
                                    divided
                 "random"         : orders them randomply
                 "angularDistance": chooses the next subset with the
                                    biggest angular distance with the
                                    ones used
    Usage
    --------
    >>> import numpy as np
    >>> import tigre
    >>> import tigre.algorithms as algs
    >>> from tigre.demos.Test_data import data_loader
    >>> geo = tigre.geometry(mode='cone',default_geo=True,
    >>>                         nVoxel=np.array([64,64,64]))
    >>> angles = np.linspace(0,2*np.pi,100)
    >>> src_img = data_loader.load_head_phantom(geo.nVoxel)
    >>> proj = tigre.Ax(src_img,geo,angles)
    >>> output = algs.iterativereconalg(proj,geo,angles,niter=50
    >>>                                 blocksize=20)

    tigre.demos.run() to launch ipython notebook file with examples.

     """

    def __init__(self, proj, geo, angles, niter, **kwargs):

        self.proj = proj
        self.angles = angles
        self.geo = geo
        self.niter = niter

        options = dict(blocksize=20, lmbda=1, lmbda_red=0.99,
                       OrderStrategy=None, Quameasopts=None,
                       init=None,verbose=True, noneg=True,
                       computel2=False, dataminimizing='art_data_minimizing',
                       name='Iterative Reconstruction', sup_kw_warning = False)
        allowed_keywords = ['V','W','log_parameters','angleblocks','angle_index','delta','regularisation']
        self.__dict__.update(options)
        self.__dict__.update(**kwargs)
        for kw in kwargs.keys():
            if not options.has_key(kw) and (kw not in allowed_keywords):
                if self.verbose:
                    if not kwargs.get('sup_kw_warning'):
                        # Note: might not want this warning (typo checking).
                        print("Warning: " + kw + " not recognised as default parameter for instance of IterativeReconAlg.")
        if self.angles.ndim == 1:
            a1 = self.angles
            a2 = np.zeros(self.angles.shape[0], dtype=np.float32)
            setattr(self, 'angles', np.vstack((a1, a2, a2)).T)
        if not all([hasattr(self, 'angleindex'), hasattr(self, 'angleblocks')]):
            self.set_angle_index()
        if not hasattr(self, 'W'):
            self.set_w()
        if not hasattr(self, 'V'):
            self.set_v()
        if not hasattr(self, 'res'):
            self.set_res()
        setattr(self, 'lq', [])  # quameasoptslist
        setattr(self, 'l2l', [])  # l2list

    def set_w(self):
        """
        Calculates value of W if this is not given.
        :return: None
        """
        geox = copy.deepcopy(self.geo)
        geox.sVoxel[0:] = self.geo.DSD - self.geo.DSO
        geox.sVoxel[2] = max(geox.sDetector[1], geox.sVoxel[2])
        geox.nVoxel = np.array([2, 2, 2])
        geox.dVoxel = geox.sVoxel / geox.nVoxel
        W = Ax(np.ones(geox.nVoxel, dtype=np.float32), geox, self.angles, "ray-voxel")
        W[W <= min(self.geo.dVoxel / 4)] = np.inf
        W = 1./W
        setattr(self, 'W', W)

    def set_v(self):
        """
        Computes value of V parameter if this is not given.
        :return: None
        """
        geo = self.geo
        if geo.mode != 'parallel':

            start = geo.sVoxel[1] / 2 - geo.dVoxel[1] / 2 + geo.offOrigin[1]
            stop = -geo.sVoxel[1] / 2 + geo.dVoxel[1] / 2 + geo.offOrigin[1]
            step = -geo.dVoxel[1]

            xv = np.arange(start, stop + step, step)

            start = geo.sVoxel[2] / 2 - geo.dVoxel[2] / 2 + geo.offOrigin[2]
            stop = -geo.sVoxel[2] / 2 + geo.dVoxel[2] / 2 + geo.offOrigin[2]
            step = -geo.dVoxel[2]

            yv = -1 * np.arange(start, stop + step, step)

            (yy, xx) = np.meshgrid(yv, xv)
            A = (self.angles[:, 0] + np.pi / 2)

            V = np.empty((self.angles.shape[0],geo.nVoxel[1], geo.nVoxel[2]))
            for i in range(self.angles.shape[0]):
                if hasattr(geo.DSO,'shape') and len(geo.DSO.shape)>=1:
                    DSO = geo.DSO[i]
                else:
                    DSO = geo.DSO
                V[i] = (DSO / (DSO + (yy * np.sin(-A[i])) - (xx * np.cos(-A[i])))) ** 2

        else:
            V = np.ones((self.angles.shape[0]), dtype=np.float32)
        if self.blocksize>1:
            v_list = [np.sum(V[self.angle_index[i]],axis=0) for i in range(len(self.angleblocks))]
            V = np.stack(v_list,0)



        V = np.array(V,dtype=np.float32)
        setattr(self, 'V', V)

    def set_res(self):
        """
        Calulates initial value for res if this is not given.
        :return: None
        """
        setattr(self, 'res', np.zeros(self.geo.nVoxel, dtype=np.float32))
        init = self.init
        verbose = self.verbose
        if init == 'multigrid':
            if verbose:
                print('init multigrid in progress...')
                print('default blocksize=1 for init_multigrid(OS_SART)')
            self.res = init_multigrid(self.proj, self.geo, self.angles, alg='SART')
            if verbose:
                print('init multigrid complete.')
        if init == 'FDK':
            self.res = FDK(self.proj, self.geo, self.angles)

        if type(init) == np.ndarray:
            if (self.geo.nVoxel == init.shape).all():

                self.res = init

            else:
                raise ValueError('wrong dimension of array for initialisation')

    def set_angle_index(self):
        """
        sets angle_index and angleblock if this is not given.
        :return: None
        """
        angleblocks, angle_index = order_subsets(self.angles, self.blocksize, self.OrderStrategy)
        setattr(self, 'angleblocks', angleblocks)
        setattr(self, 'angle_index', angle_index)

    def run_main_iter(self):
        """
        Goes through the main iteration for the given configuration.
        :return: None
        """
        Quameasopts = self.Quameasopts

        for i in range(self.niter):

            res_prev = None
            if Quameasopts is not None:
                res_prev = copy.deepcopy(self.res)
            if self.verbose:
                if i == 0:
                    print(str(self.name).upper() + ' ' + "algorithm in progress.")
                    toc = time.clock()
                if i == 1:
                    tic = time.clock()
                    print('Esitmated time until completetion (s): ' + str((self.niter - 1) * (tic - toc)))
            getattr(self, self.dataminimizing)()
            self.error_measurement(res_prev, i)
    def art_data_minimizing(self):
        """
        VERBOSE:
        >>> for j in range(angleblocks):
        >>>     angle = np.array([alpha[j]], dtype=np.float32)
        >>>     proj_err = proj[angle_index[j]] - Ax(res, geo, angle, 'ray-voxel')
        >>>     weighted_err = W[angle_index[j]] * proj_err
        >>>     backprj = Atb(weighted_err, geo, angle, 'FDK')
        >>>     weighted_backprj = 1 / V[angle_index[j]] * backprj
        >>>     res += weighted_backprj
        >>>     res[res<0]=0

        :return: None
        """
        geo = copy.deepcopy(self.geo)
        for j in range(len(self.angleblocks)):
            if self.blocksize == 1:
                angle = np.array([self.angleblocks[j]], dtype=np.float32)
            else:
                angle = self.angleblocks[j]

            if geo.offOrigin.shape[0] ==self.angles.shape[0]:
               geo.offOrigin = self.geo.offOrigin[j]
            if geo.offDetector.shape[0] == self.angles.shape[0]:
                geo.offOrin = self.geo.offDetector[j]
            if geo.rotDetector.shape[0] ==self.angles.shape[0]:
                geo.rotDetector=self.geo.rotDetector[j]
            if hasattr(geo.DSD,'shape') and len((geo.DSD.shape)):
                if geo.DSD.shape[0] ==self.angles.shape[0]:
                    geo.DSD = self.geo.DSD[j]
            if hasattr(geo.DSO,'shape') and len((geo.DSD.shape)):
                if geo.DSO.shape[0] ==self.angles.shape[0]:
                    geo.DSO = self.geo.DSO[j]

            self.gradient_descent(geo,angle,j)

            if self.noneg:
                self.res = self.res.clip(min=0)
    def third_dim_sum(self,V):
        if V.ndim == 3:
            return np.sum(V, axis=2, dtype=np.float32)
        else:
            return V

    def minimizeTV(self,res_prev,dtvg):
        return minTV(res_prev,dtvg,self.numiter_tv)

    def minimizeAwTV(self,res_prev,dtvg):
        return AwminTV(res_prev,dtvg,self.numiter_tv,self.delta)

    def error_measurement(self, res_prev, iter):
        if self.Quameasopts is not None and iter > 0:
            self.lq.append(MQ(self.res, res_prev, self.Quameasopts))
        if self.computel2:
            # compute l2 borm for b-Ax
            errornow = im3DNORM(self.proj - Ax(self.res, self.geo, self.angles, 'ray-voxel'), 2)
            self.l2l.append(errornow)

    def gradient_descent(self, geo, angle, iteration):
        self.res += self.lmbda * 1. / self.V[iteration] * Atb(self.W[self.angle_index[iteration]] * (self.proj[self.angle_index[iteration]]
                                                                                                     - Ax(self.res, geo, angle, 'interpolated')), geo, angle, 'FDK')

    def getres(self):
        return self.res

    def getl2(self):
        return self.l2l

    def __str__(self):
        parameters = []
        for item in self.__dict__:
            if item == 'geo':
                pass
            elif hasattr(self.__dict__.get(item), 'shape'):
                if self.__dict__.get(item).ravel().shape[0] > 100:
                    parameters.append(item + ' shape: ' + str(self.__dict__.get(item).shape))
            else:
                parameters.append(item + ': ' + str(self.__dict__.get(item)))

        return '\n'.join(parameters)

def decorator(IterativeReconAlg, name=None, docstring=None):
    """
    Calls run_main_iter when parameters are given to it.

    :param IterativeReconAlg: obj, class
        instance of IterativeReconAlg
    :param name: str
        for name of func
    :param docstring: str
        other documentation that may need to be included from external source.
    :return: func

    Examples
    --------
    >>> import tigre
    >>> from tigre.demos.Test_data.data_loader import load_head_phantom
    >>> geo = tigre.geometry_defaut(high_quality=False)
    >>> src = load_head_phantom(number_of_voxels=geo.nVoxel)
    >>> proj = Ax(src,geo,angles)
    >>> angles = np.linspace(0,2*np.pi,100)
    >>> iterativereconalg = decorator(IterativeReconAlg)
    >>> output = iterativereconalg(proj,geo,angles, niter=50)

    """

    def iterativereconalg(proj, geo, angles, niter, **kwargs):
        alg = IterativeReconAlg(proj, geo, angles, niter, **kwargs)
        if name is not None:
            alg.name = name
        alg.run_main_iter()
        if alg.computel2:
            return alg.getres(), alg.getl2()
        else:
            return alg.getres()

    if docstring is not None:
        setattr(iterativereconalg, '__doc__', docstring + IterativeReconAlg.__doc__)
    else:
        setattr(iterativereconalg, '__doc__', IterativeReconAlg.__doc__)
    if name is not None:
        setattr(iterativereconalg, '__name__', name)
    return iterativereconalg


iterativereconalg = decorator(IterativeReconAlg)



import os
import subprocess
import tempfile

import nbformat


def _notebook_run(path):
    """Execute a notebook via nbconvert and collect output.
       :returns (parsed nb object, execution errors)
    """
    dirname, __ = os.path.split(path)
    os.chdir(dirname)
    with tempfile.NamedTemporaryFile(suffix=".ipynb") as fout:
        args = ["nbconvert", "--to", "notebook", "--execute",
                "--ExecutePreprocessor.timeout=60",
                "--output", fout.name, path]
        subprocess.check_call(args)

        fout.seek(0)
        nb = nbformat.read(fout, nbformat.current_nbformat)

    errors = [output for cell in nb.cells if "outputs" in cell
              for output in cell["outputs"] \
              if output.output_type == "error"]

    return nb, errors


def test_ipynb():
    nb, errors = _notebook_run('../tigre/demos/d01_create_geometry.ipynb')
    assert errors == []


if __name__ == '__main__':
    test_ipynb()



from __future__ import print_function
import os
import sys

import tigre
import numpy as np
import tigre.demos.Test_data.data_loader as data_loader
import tigre.algorithms as algs
from tigre.utilities.Ax import Ax
from tigre.algorithms.iterative_recon_alg import IterativeReconAlg
from matplotlib import pyplot as plt

# ---------------------PLOT---------------------------
def plot_algs(alglist, proj, geo, angles, niter=10, **kwargs):
    l= []
    for alg in alglist:
        if alg in ['fbp','fdk','FDK']:
            l.append(getattr(algs, alg)(proj, geo, angles)[geo.nVoxel[0]/2])
        else:
            kwargs.update(dict(sup_kw_warning=True))
            l.append(getattr(algs, alg)(proj, geo, angles, niter, **kwargs)[geo.nVoxel[0]/2])
    for i in range(len(l)):
        plt.matshow(l[i])
        plt.title(str(alglist[i]) + ' ' + str(geo.mode))
        if kwargs.has_key('colorbar'):
            plt.colorbar()
    plt.show()


def input_parser(sysarglist):

    if len(sysarglist) < 2:
        raise ValueError('visual test requires at least 2 input parameters, type --help')
    if sys.argv[1]== '--help':
        print('arg 1: mode: cone, parallel \n' +
              'arg 2: alglist, eg: "ossart sirt fdk" \n'
              'arg 3: kwargs, eg: "blocksize=20, nVoxel=32,64,128')
        raise SystemExit()

    kwargs = dict(niter=10,
                  nVoxel=np.array([64, 64, 64]),
                  colorbar=True,
                  krylov='interpolated')

    kwargs.update(dict(mode=sysarglist[1],
                       alglist=sysarglist[2]))
    if len(sysarglist) > 3:
        for arg in sysarglist[3].split(' '):
            k = arg.split("=")[0]
            v = arg.split("=")[1]
            if k == 'nVoxel':
                v = np.array([int(val) for val in v.split(',')])
            kwargs.update({k: v})
            print(kwargs)

    return kwargs


def setUp(**kwargs):
    if kwargs.get('mode') == 'cone':
        geo = tigre.geometry(mode='cone', nVoxel=kwargs.get('nVoxel'), default_geo=True)
        angles_1 = np.linspace(0, 2 * np.pi, 100, dtype=np.float32)
        angles_3 = np.zeros((100), dtype=np.float32)
        angles = np.vstack((angles_1, angles_3, angles_3)).T
        img = data_loader.load_head_phantom(number_of_voxels=geo.nVoxel)
        proj = tigre.Ax(img, geo, angles, kwargs.get('krylov'))
        if kwargs.get('alglist') == 'all':
            alglist = ['sart', 'sirt', 'ossart', 'asd_pocs', 'FDK', 'cgls']
        else:
            alglist = kwargs.pop('alglist').split()

        plot_algs(alglist, proj, geo, angles, niter=int(kwargs.pop('niter')), **kwargs)

    if kwargs.get('mode') == 'parallel':
        geo = tigre.geometry(mode='parallel', nVoxel=kwargs.get('nVoxel'), default_geo=True)
        angles_1 = np.linspace(0, 2 * np.pi, 100, dtype=np.float32)
        angles_3 = np.zeros((100), dtype=np.float32)
        angles = np.vstack((angles_1, angles_3, angles_3)).T
        img = data_loader.load_head_phantom(number_of_voxels=geo.nVoxel)
        proj = tigre.Ax(img, geo, angles, kwargs.get('krylov'))
        if kwargs.get('alglist') == 'all':
            alglist = ['sart', 'sirt', 'ossart', 'asd_pocs', 'fbp', 'cgls']
        else:
            alglist = kwargs.pop('alglist').split()

        plot_algs(alglist, proj, geo, angles, niter=int(kwargs.pop('niter')), **kwargs)


if __name__ == '__main__':
    setUp(**input_parser(sys.argv))



import unittest
from tigre.utilities.geometry import geometry
from tigre.utilities.Ax import Ax
from tigre.utilities.Atb import Atb
import numpy as np

class ProjectionTestCase(unittest.TestCase):

    def setUp(self):
        geo = geometry()
        geo.DSD = 1536
        geo.DSO = 1000
        geo.nDetector = np.array((128, 127))
        geo.dDetector = np.array((0.8, 0.8)) * 4
        geo.sDetector = geo.nDetector * geo.dDetector
        geo.nVoxel = np.array((63, 62, 61))
        geo.sVoxel = np.array((256, 256, 256))
        geo.dVoxel = geo.sVoxel / geo.nVoxel
        geo.offOrigin = np.array((0, 0, 0))
        geo.offDetector = np.array((0, 0))
        geo.accuracy = 0.5
        geo.mode = 'cone'
        self.angles = np.linspace(0, 2 * np.pi, 100, dtype=np.float32)
        self.geo = geo
        self.img = np.ones((63, 62, 61), dtype=np.float32)
        self.proj = Ax(self.img,self.geo,self.angles)
    def test_geo(self):
        self.assertIsNone(self.geo.check_geo(self.angles))

    def test_proj_interpolated_shape(self):
        img = self.img
        self.assertTupleEqual(Ax(img,self.geo,self.angles,'interpolated').shape,(100,128,127))

    def test_proj_ray_voxel_shape(self):
        self.assertTupleEqual(Ax(self.img,self.geo,self.angles,'ray-voxel').shape,(100,128,127))

    def test_proj_parallel_interpolated_shape(self):
        setattr(self.geo,'mode','parallel')
        self.assertTupleEqual(Ax(self.img,self.geo,self.angles,'interpolated').shape,(100,128,127))

    def test_proj_parallel_ray_voxel_shape(self):
        setattr(self.geo,'mode','parallel')
        self.assertTupleEqual(Ax(self.img,self.geo,self.angles,'ray-voxel').shape,(100,128,127))

    def test_backproj_matched_cone_shape(self):
        setattr(self.geo,'mode','cone')
        self.assertTupleEqual(Atb(self.proj,self.geo,self.angles,'matched').shape,(63,62,61))

    def test_backproj_FDK_cone_shape(self):
        setattr(self.geo,'mode','cone')
        self.assertTupleEqual(Atb(self.proj, self.geo, self.angles, 'FDK').shape, (63, 62, 61))

    def test_backproj_parallel_shape(self):
        setattr(self.geo,'mode','parallel')
        self.assertTupleEqual(Atb(self.proj, self.geo, self.angles,'matched').shape, (63, 62, 61))

    def test_backproj_matched_cone_spherical_shape(self):
        setattr(self.geo,'mode','cone')
        angles_1 = self.angles
        angles_2 = np.ones((len(angles_1)),dtype=np.float32)*np.array([np.pi/4],dtype=np.float32)
        angles_3 = np.zeros((len(angles_1)),dtype=np.float32)
        new_angles = np.vstack((angles_1,angles_2,angles_3)).T
        self.assertTupleEqual(Atb(self.proj, self.geo, new_angles,'matched').shape, (63, 62, 61))

    def test_backproj_FDK_cone_spherical_shape(self):
        setattr(self.geo,'mode','cone')
        angles_1 = self.angles
        angles_2 = np.ones((len(angles_1)), dtype=np.float32) * np.array([np.pi / 4], dtype=np.float32)
        angles_3 = np.zeros((len(angles_1)), dtype=np.float32)
        new_angles = np.vstack((angles_1, angles_2, angles_3)).T
        self.assertTupleEqual(Atb(self.proj, self.geo, new_angles,'FDK').shape, (63, 62, 61))

    def test_backproj_parallel_spherical_shape(self):
        setattr(self.geo,'mode','parallel')
        angles_1 = self.angles
        angles_2 = np.ones((len(angles_1)), dtype=np.float32) * np.array([np.pi / 4], dtype=np.float32)
        angles_3 = np.zeros((len(angles_1)), dtype=np.float32)
        new_angles = np.vstack((angles_1, angles_2, angles_3)).T
        self.assertTupleEqual(Atb(self.proj, self.geo, new_angles).shape, (63, 62, 61))

if __name__=='__main__':
        unittest.main()



from __future__ import print_function
from __future__ import division
from tigre.utilities.parkerweight import parkerweight
import numpy as np

import warnings

#TODO: Fix parker
def filtering(proj,geo,angles,parker,verbose=False):
    if parker:
        proj=parkerweight(proj.transpose(0,2,1),geo,angles,parker).transpose(0,2,1)
        # proj=parkerweight(proj,geo,angles,parker)
    filt_len=max(64,2**nextpow2(2*max(geo.nDetector)))
    ramp_kernel=ramp_flat(filt_len)

    d=1
    filt=filter(geo.filter,ramp_kernel[0],filt_len,d,verbose=verbose)
    filt=np.kron(np.ones((np.int64(geo.nDetector[0]),1)),filt)
    for i in range(angles.shape[0]):
        fproj=np.zeros((geo.nDetector[0],filt_len),dtype=np.float32)
        fproj[:,int(filt_len/2-geo.nDetector[1]/2):int(filt_len/2+geo.nDetector[1]/2)]=proj[i]
        fproj=np.fft.fft(fproj,axis=1)

        fproj=fproj*filt

        fproj=np.real(np.fft.ifft(fproj,axis=1))
        proj[i]=fproj[:,int(filt_len/2-geo.nDetector[1]/2):int(filt_len/2+geo.nDetector[1]/2)]/2/geo.dDetector[0]*(2*np.pi/
                                                                                                 len(angles)
                                                                                                 )/2*(geo.DSD[0]/geo.DSO[0])

    return proj

def ramp_flat(n,verbose=False):
    nn=np.arange(-n/2,n/2)
    h=np.zeros(nn.shape,dtype=np.float32)
    h[int(n/2)]=1/4
    odd=nn%2==1
    h[odd]=-1/(np.pi*nn[odd])**2
    return h, nn

def filter(filter,kernel,order,d,verbose=False):
    f_kernel=abs(np.fft.fft(kernel))*2

    filt=f_kernel[:int((order/2)+1)]
    w=2*np.pi*np.arange(len(filt))/order
    if filter not in ['ram_lak','shepp_logan','cosine','hamming','hann',None]:
        raise ValueError('filter not recognised: '+str(filter))

    if filter in {'ram_lak', None}:
        if filter is None:
            if verbose:
                warnings.warn('no filter selected, using default ram_lak')
        pass
    if filter=='shepp_logan':
        filt[1:]*=(np.sin(
                                w[1:]/(2*d))/(w[1:]/(2*d)
                                              )
                           )
    if filter=='cosine':
        filt[1:]*=np.cos(w[1:]/(2*d))
    if filter=='hamming':
        filt[1:]*=(.54+.46*np.cos(w[1:]/d))
    if filter =='hann':
        filt[1:]*=(1+np.cos(w[1:])/d)/2

    filt[w>np.pi*d]=0
    filt=np.hstack((filt,filt[1:-1][::-1]))
    return filt


def nextpow2(n):
    i=1
    while (2**i)<n:
        i+=1
    return i



from __future__ import print_function
import tigre
import sys, traceback
import os
import numpy as np
from tigre.demos.Test_data import data_loader
from matplotlib import pyplot as plt
import time
import tigre.algorithms as algs
import warnings
from matplotlib import pyplot as plt

warnings.filterwarnings("error")
from git import Repo


def testandlog(alglist, geo, angles, niter, saveresult=False, newsubdirectory=False, **kwargs):
    nangles = angles.shape[0]
    # createlogfile

    dirname = os.path.dirname(__file__)
    kwargs.update(dict(blocksize=20))
    # create top level folder if this does not exist

    if 'logs' not in os.listdir(dirname):
        os.system('mkdir ' + 'logs')
    logdirectory = os.path.join(dirname, 'logs')
    logdate = time.strftime("%a_%b_%d")
    if logdate not in os.listdir(logdirectory):
        os.system('mkdir ' + os.path.join(logdirectory, logdate))
    subdirectory = os.path.join(logdirectory, logdate)
    if 'subdirectoryname' in kwargs:
        subdirectoryname = kwargs['subdirectoryname']
    else:
        subdirectoryname = 'test' + str(len(os.listdir(subdirectory)) - 1)

    if newsubdirectory:
        subdirectoryname = 'test' + str(len(os.listdir(subdirectory)))

    # create subdirectory for test if this does not exist

    if subdirectoryname not in os.listdir(subdirectory):
        os.system('mkdir ' + os.path.join(subdirectory, subdirectoryname))
    subdirectory = os.path.join(logdirectory, logdate, subdirectoryname)
    timestamp = str(time.asctime()).replace(' ', '_')

    # create/append to log file

    logfilename = None
    for filename in os.listdir(subdirectory):
        if os.path.splitext(filename)[-1] == '.log':
            logfilename = os.path.split(filename)[-1]
    logflist = []
    if logfilename is not None:
        logflist.extend(open(os.path.join(subdirectory, logfilename), 'r').readlines())
    else:
        logfilename = timestamp + '.log'
    try:
        algsuccess = np.load(os.path.join(subdirectory, os.path.splitext(logfilename)[0]) + '.npy').item()
    except Exception:
        algsuccess = dict()
    # TODO: need to fix directory path here.
    repo_path = os.path.join(dirname, '..', '..', '.git/')
    repo = Repo(repo_path)
    commit = list(repo.iter_commits(repo.active_branch))[0]
    uname = str(os.uname())
    current_config = dict(uname=uname, commithash=commit.hexsha)
    try:
        prev_config = open(os.path.join(subdirectory, os.path.splitext(logfilename)[0] + '_config') + '.npy').item()
    except Exception:
        prev_config = dict()
        prev_config.update(current_config)

    for item in current_config:
        if prev_config[item] != current_config[item]:
            logflist.append('------------------------------------------------\n')
            logflist.append('Configuration changed for' + item + ':\n')
            logflist.append('From: ' + str(prev_config[item]))
            logflist.append('To  : ' + str(current_config[item]))

    if len(logflist) == 0:  # If file is empty
        # MACHINE

        logflist.append(uname + '\n')

        # GIT header for logfile

        if not repo.bare:
            logflist.append('------------------------------------------------\n')
            logflist.append(make_repository_str(repo))

            # create list of commits then print some of them to stdout

            logflist.append(make_commit_str(commit))
            logflist.append('\n------------------------------------------------\n')
        else:
            print('Could not load repository at {} :('.format(repo_path))

        logflist.append('GEOMETRY used for instance of testandlog: \n')
        for item in geo.__dict__:
            logflist.append(item + ': ' + str(getattr(geo, item)) + '\n')
        logflist.append('------------------------------------------------\n')

    # Run the algorithms

    source_img = data_loader.load_head_phantom(number_of_voxels=geo.nVoxel)
    proj = tigre.Ax(source_img, geo, angles)

    for alg in alglist:
        logflist.append(str(alg).upper() + ' ' + time.asctime() + '\n')
        logflist.append('nproj: ' + str(nangles) + ' niter: ' + str(niter) + '\n')
        if np.sum(angles[:, 1]) != 0 or np.sum(angles[:, 2]) != 0:
            spherical = True
        else:
            spherical = False
        logflist.append('spherical projection: ' + str(spherical) + '\n')
        algpassed = False
        try:
            tic = time.clock()
            res = getattr(algs, alg)(proj, geo, angles, niter=niter, **kwargs)
            algpassed = True
        except Exception:
            formatedlines = traceback.format_exc()
            logflist.append('ERROR at ' + str(time.strftime("%H:%M:%S") + '\n'))
            logflist.append(''.join(formatedlines) + '\n')
        finally:
            toc = time.clock()
            algsuccess.update({alg: algpassed})
            np.save(os.path.join(subdirectory, os.path.splitext(logfilename)[0]), algsuccess)
            np.save(os.path.join(subdirectory,os.path.splitext(logfilename)[0])+'_config',current_config)
            pass
        if algpassed:
            logflist.append('total time taken: ' + str(abs(tic - toc)) + '\n')
        logflist.append('------------------------------------------------\n')
        logf = open(os.path.join(subdirectory, logfilename), 'w')
        logf.write(''.join(logflist))
        logf.close()
        if saveresult and algpassed:
            warnings.filterwarnings("ignore")
            plt.figure()
            plt.subplot(3, 1, 1)
            plt.imshow(res[geo.nVoxel[0] / 2])
            plt.title('results for ' + alg)
            plt.ylabel('dim 0')

            plt.subplot(3, 1, 2)
            plt.imshow(res[:, geo.nVoxel[1] / 2])
            plt.ylabel('dim 1')

            plt.subplot(3, 1, 3)
            plt.imshow(res[:, :, geo.nVoxel[2] / 2])
            plt.ylabel('dim 2')
            plt.savefig(os.path.join(subdirectory, alg + '_' + geo.mode + '_' + timestamp + '.png'))
            warnings.filterwarnings("error")


def make_commit_str(commit):
    l = []
    l.append(str(commit.hexsha))
    l.append("\"{}\" by {} ({})".format(commit.summary,
                                        commit.author.name,
                                        commit.author.email))
    l.append(str(commit.authored_datetime))
    l.append(str("count: {} and size: {}".format(commit.count(),
                                                 commit.size)))
    commit_str = ('\n'.join(l))
    return commit_str


def make_repository_str(repo):
    l = []
    l.append('Repo description: {}'.format(repo.description))
    l.append('Repo active branch is {}'.format(repo.active_branch))
    for remote in repo.remotes:
        l.append('Remote named "{}" with URL "{}"'.format(remote, remote.url))
    l.append('Last commit for repo is {}.'.format(str(repo.head.commit.hexsha)))
    repo_str = ('\n'.join(l))
    return repo_str


def log_summary(logdate, tobool=False):
    dirname = os.path.dirname(__file__)
    if 'logs' not in os.listdir(dirname):
        raise Exception('Logs not found under ' + dirname)
    logdirectory = os.path.join(dirname, 'logs')

    if logdate not in os.listdir(logdirectory):
        raise Exception('Date directory ' + logdate + ' not found.')
    logdatedirectory = os.path.join(logdirectory, logdate)
    logflist = []
    logfilename = ''
    for filename in os.listdir(logdirectory):
        if os.path.splitext(filename)[-1] == '.log':
            logfilename = os.path.split(filename)[-1]

    if logdate in logfilename:
        logflist.extend(open(os.path.join(logdirectory, logfilename), 'r').readlines())
    else:
        logfilename = logdate + '_testsummary.log'
        logflist.append(str(os.uname()) + '\n')
        logflist.append('------------------------------------------------\n')
        repo_path = os.path.join(dirname, '..', '..', '.git/')
        repo = Repo(repo_path)
        if not repo.bare:
            logflist.append('------------------------------------------------\n')
            logflist.append(make_repository_str(repo))

            # create list of commits then print some of them to stdout
            commit = list(repo.iter_commits(repo.active_branch))[0]
            logflist.append(make_commit_str(commit))
            logflist.append('\n------------------------------------------------\n')
        else:
            print('Could not load repository at {} :('.format(repo_path))

    toboolist = []
    logflist.append('summary carried out on ' + time.asctime() + '\n')
    for subdirectory in os.listdir(logdatedirectory):
        for testfile in os.listdir(os.path.join(logdatedirectory, subdirectory)):
            if os.path.splitext(testfile)[-1] == '.log':
                timestamp = os.path.splitext(testfile)[0]

                try:
                    testpassed = np.load(os.path.join(logdatedirectory, subdirectory, timestamp) + '.npy').item()
                    testconfig = np.load(os.path.join(logdatedirectory,subdirectory,timestamp)+'_config.npy').item()
                    formattedlist = []
                    for alg in testpassed:
                        toboolist.append(testpassed[alg])
                        if testpassed[alg]:
                            formattedlist.append('      ' + alg + ' : PASSED')
                        else:
                            formattedlist.append('      ' + alg + ' : ERROR')
                    logflist.append('\n------------------------------------------------\n')
                    logflist.append(subdirectory + ': \n' + str('\n'.join(formattedlist)) + '\n')
                    logflist.extend([item + ': ' + str(testconfig[item]) + '\n' for item in testconfig])
                except IOError:
                    logflist.append(subdirectory + ' no .npy file found. \n')

    logflist.append('\n All tests passed:' + str(all(toboolist)) + '\n')
    logflist.append('\n------------------------------------------------\n')

    with open(os.path.join(logdirectory, logfilename), 'w') as logf:
        logf.write(''.join(logflist))

    if tobool:
        return all(toboolist)


if __name__ == '__main__':
    method_name = sys.argv[1]
    param_name = sys.argv[2]
    tobool = False
    if len(sys.argv) == 4:
        tobool = sys.argv[3]

    getattr(sys.modules[__name__], method_name)(param_name, tobool)



from __future__ import division
import unittest
import tigre
from tigre.utilities.Ax import Ax
import numpy as np
from tigre.algorithms.iterative_recon_alg import IterativeReconAlg
import tigre.algorithms as algs
class AlgorithmsTestCase(unittest.TestCase):

    def setUp(self):
        geo = tigre.geometry()
        geo.DSD = 1536
        geo.DSO = 1000
        geo.nDetector = np.array((128, 127))
        geo.dDetector = np.array((0.8, 0.8)) * 4.
        geo.sDetector = geo.nDetector * geo.dDetector
        geo.nVoxel = np.array((63, 62, 61))
        geo.sVoxel = np.array((256, 256, 256))
        geo.dVoxel = geo.sVoxel / geo.nVoxel
        geo.offOrigin = np.array((0, 0, 0))
        geo.offDetector = np.array((0, 0))
        geo.accuracy = 0.5
        geo.mode = 'parallel'
        self.angles = np.linspace(0, 2 * np.pi, 100, dtype=np.float32)
        self.geo = geo
        self.img = np.ones((63, 62, 61), dtype=np.float32)
        self.proj = Ax(self.img,self.geo,self.angles)
        self.niter=2
    def test_mainiter_itrecalg(self):
        alg = IterativeReconAlg(self.proj,self.geo,self.angles,self.niter,**dict(verbose=False))
        self.assertIsNone(alg.run_main_iter())

    def test_shape_sart(self):
        self.assertTupleEqual(tigre.algorithms.sart(self.proj,self.geo,self.angles,niter=1).shape,
                              tuple(self.geo.nVoxel))
    def test_shape_sirt(self):
        self.assertTupleEqual(tigre.algorithms.sirt(self.proj,self.geo,self.angles,niter=1).shape,
                              tuple(self.geo.nVoxel))
    def test_shape_ossart(self):
        self.assertTupleEqual(tigre.algorithms.sirt(self.proj,self.geo,self.angles,niter=1).shape,
                              tuple(self.geo.nVoxel))
    def test_shape_asdpocs(self):
        self.assertTupleEqual(algs.asd_pocs(self.proj,self.geo,self.angles,niter=1).shape,
                              tuple(self.geo.nVoxel))
    def test_shape_awasdpocs(self):
        self.assertTupleEqual(algs.awasd_pocs(self.proj,self.geo,self.angles,niter=1).shape,
                              tuple(self.geo.nVoxel))
if __name__=='__main__':
        unittest.main()


cimport numpy as np
import numpy as np
from tigre.Source._types cimport Geometry as c_Geometry, convert_to_c_geometry, free_c_geometry

# Numpy must be initialized. When using numpy from C or Cython you must
# _always_ do that, or you will have segfaults
np.import_array()

from libc.stdlib cimport malloc, free

cdef extern from "numpy/arrayobject.h":
    void PyArray_ENABLEFLAGS(np.ndarray arr, int flags)
    void PyArray_CLEARFLAGS(np.ndarray arr, int flags)


cdef extern from "Siddon_projection.hpp":
    cdef int siddon_ray_projection(float* img, c_Geometry geo, float** result, float* alphas, int nalpha)
cdef extern from "Siddon_projection_parallel.hpp":
    cdef int siddon_ray_projection_parallel(float* img, c_Geometry geo, float** result, float* alphas, int nalpha)
cdef extern from "ray_interpolated_projection.hpp":
    cdef int interpolation_projection(float* img, c_Geometry geo, float** result, float* alphas, int nalpha)
cdef extern from "ray_interpolated_projection_parallel.hpp":
    cdef int interpolation_projection_parallel(float* img, c_Geometry geo, float** result, float* alphas, int nalpha)

def cuda_raise_errors(error_code):
    if error_code:
        raise ValueError('TIGRE: Call to Ax failed')

def _Ax_ext(np.ndarray[np.float32_t, ndim=3] img, geometry, np.ndarray[np.float32_t, ndim=2] angles, projection_type="ray-voxel", mode="cone"):

    cdef int total_projections = angles.shape[0]

    # TODO: For now we will just make a new geometry (C) struct from the python one,
    # but this is really ugly and should be changed.

    #PERMUTE INPUT: convert_contig_mode(C_CONTIG) -> F_CONTIG
    # (N, V, U) -> (U, V, N) 
    geometry.convert_contig_mode()
    cdef c_Geometry* c_geometry = convert_to_c_geometry(geometry, total_projections)

    cdef float** c_projections = <float**> malloc(total_projections * sizeof(float*))

    cdef int i = 0
    for i in range(total_projections):
       c_projections[i] = <float*> malloc(geometry.nDetector[0] * geometry.nDetector[1] * sizeof(float))

    cdef float* c_angles = <float*> angles.data

    if projection_type == "ray-voxel":
        interpolated = False
    elif projection_type == "interpolated":
        interpolated = True
    else:
        print("Warning: Unknown projection_type, using default ray-voxel")
        interpolated = False

    if mode == "parallel":
        cone_beam = False
    elif mode == "cone":
        cone_beam = True
    else:
        print("Warning: Unknown mode, using default cone beam")
        cone_beam = True
    #PERMUTE INPUT: (Z, Y, X) -> (X ,Y ,Z)
    img = img.transpose().copy(order='F')

    cdef float* c_img = <float*> img.data
    if cone_beam:
        if not interpolated:
            cuda_raise_errors(siddon_ray_projection(c_img, c_geometry[0], c_projections, c_angles, total_projections))
        else:
            cuda_raise_errors(interpolation_projection(c_img, c_geometry[0], c_projections, c_angles, total_projections))
    else:
        if not interpolated:
            cuda_raise_errors(siddon_ray_projection_parallel(c_img, c_geometry[0], c_projections, c_angles, total_projections))
        else:
            cuda_raise_errors(interpolation_projection_parallel(c_img, c_geometry[0], c_projections, c_angles, total_projections))
    cdef float theta,psi;
    theta=0;
    psi=0;
    for i in range(total_projections):
        theta+=abs(c_angles[i*3+1])
        psi  +=abs(c_angles[i*3+2])
    
    standard_rotation = True;
    
    if psi == 0.0 and theta == 0.0:
        standard_rotation=True
    else:
        standard_rotation=False

    img = img.copy(order='C')


    cdef np.npy_intp shape[2]
    shape[0] = <np.npy_intp> geometry.nDetector[0]
    shape[1] = <np.npy_intp> geometry.nDetector[1]

    projections = [None] * total_projections
    for i in range(total_projections):
        projections[i] = np.PyArray_SimpleNewFromData(2, shape, np.NPY_FLOAT32, c_projections[i])
        PyArray_ENABLEFLAGS(projections[i], np.NPY_OWNDATA)  # Attribute new memory owner


    free(c_projections)  # Free pointer array, not actual data
    free_c_geometry(c_geometry)
    geometry.convert_contig_mode()
    #TODO: check stack doesn't cause memory leak, as it seems projections needs multiple free calls
    # A possible solution is to rewrite SiddonProjection to allow 1d array results

    # PERMUTE OUTPUT: (U, V, N) -> (N, V, U)
    return np.stack(projections,0).swapaxes(1,2).copy(order='C')


from scipy.interpolate import RegularGridInterpolator as rgi
from numpy.lib.stride_tricks import as_strided
import numpy as np
import tigre
def init_multigrid(proj, geo, alpha,alg):
    # WARNING: This takes a lot of memory!
    if alg=='SART':
        italg = tigre.algorithms.sart
    if alg=='SIRT':
        italg = tigre.algorithms.sirt
    finalsize = geo.nVoxel

    maxval= max(proj.ravel())
    minval = min(proj.ravel())


    # Start with 16 (raise this for larger images)
    geo.nVoxel = np.array([16, 16, 16])
    geo.dVoxel = geo.sVoxel / geo.nVoxel
    if (geo.nVoxel > finalsize).all():
        return np.zeros(finalsize, dtype=np.float32)
    niter = 100
    initres = np.zeros(geo.nVoxel, dtype=np.float32)
    while (geo.nVoxel != finalsize).all():
        geo.dVoxel = geo.sVoxel / geo.nVoxel

        initres = italg(proj, geo, alpha, niter, init=initres,verbose=False)

        # get new dims(should be a way to do this more efficiently).

        geo.nVoxel = geo.nVoxel * 2
        geo.nVoxel[geo.nVoxel > finalsize] = finalsize[geo.nVoxel > finalsize]
        geo.dVoxel = geo.sVoxel / geo.nVoxel
        (x,y,z)=(
        np.linspace(minval, maxval, geo.nVoxel[0]/2,dtype=np.float32),
        np.linspace(minval, maxval, geo.nVoxel[1]/2,dtype=np.float32),
        np.linspace(minval, maxval, geo.nVoxel[2]/2,dtype=np.float32)
        )


        # evaluate the function sart at the points xv,yv,zv

        xv,yv,zv=[tile_array(tile_array(x,2), geo.nVoxel[0]**2),
        tile_array(tile_array(y,2), geo.nVoxel[0]**2),
        tile_array(tile_array(x,2), geo.nVoxel[0]**2)]




        initres = rgi((x, y, z), initres)(np.column_stack((xv,yv,zv)))
        initres = initres.reshape(geo.nVoxel)



    return initres
def tile_array(mat, b1):
    r, = mat.shape
    rs, = mat.strides
    x = as_strided(mat, (r, b1), (rs, 0))
    return x.reshape(r*b1)



from __future__ import division
import numpy as np


class TIGREParameters:

    def __init__(self, high_quality=True):
        if high_quality:
            # VARIABLE                                          DESCRIPTION                    UNITS
            # -------------------------------------------------------------------------------------
            self.DSD = 1536                                     # Distance Source Detector      (mm)
            self.DSO = 1000                                     # Distance Source Origin        (mm)
            # Detector parameters
            self.nDetector = np.array((512, 512))               # number of pixels              (px)
            self.dDetector = np.array((0.8, 0.8))               # size of each pixel            (mm)
            self.sDetector = self.nDetector * self.dDetector    # total size of the detector    (mm)
            # Image parameters
            self.nVoxel = np.array((256, 256, 256))             # number of voxels              (vx)
            self.sVoxel = np.array((256, 256, 256))             # total size of the image       (mm)
            self.dVoxel = self.sVoxel/self.nVoxel               # size of each voxel            (mm)
            # Offsets
            self.offOrigin = np.array((0, 0, 0))                # Offset of image from origin   (mm)
            self.offDetector = np.array((0, 0))                 # Offset of Detector            (mm)

            # Auxiliary
            self.accuracy = 0.5                                 # Accuracy of FWD proj          (vx/sample)
            # Mode
            self.mode = 'cone'                                  # parallel, cone                ...
        else:
            # VARIABLE                                          DESCRIPTION                    UNITS
            # -------------------------------------------------------------------------------------
            self.DSD = 1536                                     # Distance Source Detector      (mm)
            self.DSO = 1000                                     # Distance Source Origin        (mm)
            # Detector parameters
            self.nDetector = np.array((128, 128))             # number of pixels              (px)
            self.dDetector = np.array((0.8, 0.8))*4             # size of each pixel            (mm)
            self.sDetector = self.nDetector * self.dDetector    # total size of the detector    (mm)
            # Image parameters
            self.nVoxel = np.array((64, 100, 200))           # number of voxels              (vx)
            self.sVoxel = np.array((256, 256, 256))             # total size of the image       (mm)
            self.dVoxel = self.sVoxel / self.nVoxel             # size of each voxel            (mm)
            # Offsets
            self.offOrigin = np.array((0, 0, 0))                # Offset of image from origin   (mm)
            self.offDetector = np.array((0, 0))                 # Offset of Detector            (mm)

            # Auxiliary
            self.accuracy = 0.5                                 # Accuracy of FWD proj          (vx/sample)
            # Mode
            self.mode=None                                      # parallel, cone                ...
            self.filter=None
    def __str__(self):
        parameters = []
        parameters.append("TIGRE parameters")
        parameters.append("-----")
        parameters.append("Distance from source to detector = " + str(self.DSD) + "mm")
        parameters.append("Distance from source to origin = " + str(self.DSO) + "mm")

        parameters.append("-----")
        parameters.append("Detector parameters")
        parameters.append("Number of pixels = " + str(self.nDetector))
        parameters.append("Size of each pixel = " + str(self.dDetector) + "mm")
        parameters.append("Total size of the detector = " + str(self.sDetector) + "mm")

        parameters.append("-----")
        parameters.append("Image parameters")
        parameters.append("Number of voxels = " + str(self.nVoxel))
        parameters.append("Total size of the image = " + str(self.sVoxel) + "mm")
        parameters.append("Size of each voxel = " + str(self.dVoxel) + "mm")

        parameters.append("-----")
        parameters.append("Offset correction parameters")
        parameters.append("Offset of image from origin = " + str(self.offOrigin) + "mm")
        parameters.append("Offset of detector = " + str(self.offDetector) + "mm")

        parameters.append("-----")
        parameters.append("Auxillary parameters")
        parameters.append("Accuracy of forward projection = " + str(self.accuracy))

        return '\n'.join(parameters)



import traceback


class Error(Exception):
    """Base class for exceptions in this module"""
    pass


class TigreCudaCallError(Error):
    """Error list indexed against values in errors.hpp."""

    def __init__(self, txt, error_index):
        self.error_list = [
            "GPU: Cuda error with exit code 1.",

            "There are no available device(s) that support CUDA",

            "GPU:One (or more) of your GPUs is being heavily used by another program (possibly graphics-based). \n "
            "Free the GPU to run TIGRE",

            "GPU: Assertion Failed. Logic behind spliting flawed! Please tell: ander.biguri@gmail.com",

            "GPU: Assertion Failed. Memory needed calculation broken! Please tell: ander.biguri@gmail.com"
        ]
        self.error_index = error_index
        self.txt = txt
    def __str__(self):
        return([self.txt, self.error_list[self.error_index-1]])



